#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create lambda functions"""

import logging
import zipfile

import boto3
from tryagain import retries

from ..exceptions import RequiredKeyNotFound
from ..utils import get_details, get_lambda_arn, get_properties, get_role_arn, get_security_group_id, get_subnets

LOG = logging.getLogger(__name__)


class LambdaFunction:
    """Manipulate Lambda function."""

    def __init__(self, app, env, region, prop_path):
        """Lambda function object.

        Args:
            app (str): Application name
            env (str): Environment/Account
            region (str): AWS Region
            prop_path (str): Path of environment property file
        """
        self.app_name = app
        self.env = env
        self.region = region
        self.properties = get_properties(prop_path)
        generated = get_details(app=self.app_name)
        self.group = generated.data['project']

        try:
            self.pipeline = self.properties['pipeline']['lambda']
        except KeyError:
            raise RequiredKeyNotFound("Lambda key in pipeline.json is required.")

        self.runtime = self.pipeline['runtime']
        self.description = self.pipeline['app_description']
        self.handler = self.pipeline['handler']
        self.vpc_enabled = self.pipeline['vpc_enabled']

        self.settings = get_properties(prop_path, env=self.env, region=self.region)
        app = self.settings['app']
        self.custom_tags = app['custom_tags']
        self.lambda_environment = app['lambda_environment']
        self.lambda_layers = app['lambda_layers']
        self.lambda_destinations = app['lambda_destinations']
        self.lambda_dlq = app['lambda_dlq']
        self.lambda_filesystems = app['lambda_filesystems']
        self.lambda_tracing = app['lambda_tracing']
        self.lambda_provisioned_throughput = app['lambda_provisioned_throughput']
        self.memory = app['lambda_memory']
        self.role = app.get('lambda_role') or generated.iam()['lambda_role']
        self.timeout = app['lambda_timeout']
        self.concurrency_limit = app.get('lambda_concurrency_limit')
        self.subnet_count = app['lambda_subnet_count']

        self.role_arn = get_role_arn(self.role, self.env, self.region)

        self.session = boto3.Session(profile_name=self.env, region_name=self.region)
        self.lambda_client = self.session.client('lambda')

    def _check_lambda(self):
        """Check if lambda function exists.

        Returns:
            True if function does exist
            False if function does not exist
        """
        exists = False
        try:
            self.lambda_client.get_function(FunctionName=self.app_name)
            exists = True
        except boto3.exceptions.botocore.exceptions.ClientError:
            pass
        return exists

    def _check_lambda_alias(self):
        """Check if lambda alias exists.

        Returns:
            True if alias exists
            False if alias does not exist
        """
        aliases = self.lambda_client.list_aliases(FunctionName=self.app_name)

        matched_alias = False
        for alias in aliases['Aliases']:
            if alias['Name'] == self.env:
                LOG.info('Found alias %s for function %s', self.env, self.app_name)
                matched_alias = True
                break
        else:
            LOG.info('No alias %s found for function %s', self.env, self.app_name)
        return matched_alias

    def _vpc_config(self):
        """Get VPC config."""
        if self.vpc_enabled:
            subnets = get_subnets(env=self.env, region=self.region, purpose='internal')['subnet_ids'][self.region]
            if self.subnet_count:
                subnets = subnets[:self.subnet_count]
                LOG.info('Subnet Count of %s specified. Limiting to subnets: %s', self.subnet_count, subnets)

            security_groups = self._get_sg_ids()

            vpc_config = {'SubnetIds': subnets, 'SecurityGroupIds': security_groups}
        else:
            vpc_config = {'SubnetIds': [], 'SecurityGroupIds': []}
        LOG.debug("Lambda VPC config setup: %s", vpc_config)
        return vpc_config

    def _get_sg_ids(self):
        """Get IDs for all defined security groups.

        Returns:
            list: security group IDs for all lambda_extras
        """
        try:
            lambda_extras = self.settings['security_groups']['lambda_extras']
        except KeyError:
            lambda_extras = []

        security_groups = [self.app_name] + lambda_extras
        sg_ids = []
        for security_group in security_groups:
            sg_id = get_security_group_id(name=security_group, env=self.env, region=self.region)
            sg_ids.append(sg_id)
        return sg_ids

    @retries(max_attempts=3, wait=1, exceptions=(boto3.exceptions.botocore.exceptions.ClientError))
    def create_alias(self):
        """Create lambda alias with env name and points it to $LATEST."""
        LOG.info('Creating alias %s', self.env)

        try:
            self.lambda_client.create_alias(
                FunctionName=self.app_name,
                Name=self.env,
                FunctionVersion='$LATEST',
                Description='Alias for {}'.format(self.env))
        except boto3.exceptions.botocore.exceptions.ClientError as error:
            LOG.debug('Create alias error: %s', error)
            LOG.info("Alias creation failed. Retrying...")
            raise

    @retries(max_attempts=3, wait=1, exceptions=(boto3.exceptions.botocore.exceptions.ClientError))
    def update_alias(self):
        """Update lambda alias to point to $LATEST."""
        LOG.info('Updating alias %s to point to $LATEST', self.env)

        try:
            self.lambda_client.update_alias(FunctionName=self.app_name, Name=self.env, FunctionVersion='$LATEST')
        except boto3.exceptions.botocore.exceptions.ClientError as error:
            LOG.debug('Update alias error: %s', error)
            LOG.info("Alias update failed. Retrying...")
            raise

    @retries(max_attempts=3, wait=1, exceptions=(SystemExit))
    def update_function_configuration(self, vpc_config):
        """Update existing Lambda function configuration.

        Args:
            vpc_config (dict): Dictionary of SubnetIds and SecurityGroupsIds for using
                               a VPC in lambda
        """
        LOG.info('Updating configuration for lambda function: %s', self.app_name)

        default_tags = {'app_group': self.group, 'app_name': self.app_name}
        lambda_tags = {**default_tags, **self.custom_tags}

        try:
            self.lambda_client.update_function_configuration(
                Environment=self.lambda_environment,
                FunctionName=self.app_name,
                Runtime=self.runtime,
                Role=self.role_arn,
                Handler=self.handler,
                Description=self.description,
                Timeout=int(self.timeout),
                MemorySize=int(self.memory),
                VpcConfig=vpc_config,
                Layers=self.lambda_layers,
                DeadLetterConfig=self.lambda_dlq,
                TracingConfig=self.lambda_tracing,
                FileSystemConfigs=self.lambda_filesystems)

            if self.concurrency_limit:
                self.lambda_client.put_function_concurrency(
                    FunctionName=self.app_name,
                    ReservedConcurrentExecutions=self.concurrency_limit
                )
            else:
                self.lambda_client.delete_function_concurrency(FunctionName=self.app_name)

            if self.lambda_destinations:
                self.lambda_client.put_function_event_invoke_config(
                    FunctionName=self.app_name,
                    DestinationConfig=self.lambda_destinations
                )

            if self.lambda_provisioned_throughput:
                self.lambda_client.put_provisioned_concurrency_config(
                    FuctionName=self.app_name,
                    Qualifier=self.env,
                    ProvisionedConcurrentExecutions=self.lambda_provisioned_throughput)
            else:
                self.lambda_client.delete_provisioned_concurrency_config(
                    FunctionName=self.app_name,
                    Qualifier=self.env)

        except boto3.exceptions.botocore.exceptions.ClientError as error:
            if 'CreateNetworkInterface' in error.response['Error']['Message']:
                message = '{0} is missing "ec2:CreateNetworkInterface"'.format(self.role_arn)
                LOG.debug(message)
                raise SystemExit(message)

            raise
        LOG.info('Updating Lambda function tags')

        lambda_arn = get_lambda_arn(self.app_name, self.env, self.region)
        self.lambda_client.tag_resource(Resource=lambda_arn, Tags=lambda_tags)

        LOG.info("Successfully updated Lambda configuration.")

    @retries(max_attempts=3, wait=1, exceptions=(SystemExit))
    def create_function(self, vpc_config):
        """Create lambda function, configures lambda parameters.

        We need to upload non-zero zip when creating function. Uploading
        hello_world python lambda function since AWS doesn't care which
        executable is in ZIP.

        Args:
            vpc_config (dict): Dictionary of SubnetIds and SecurityGroupsIds for using
                               a VPC in lambda
        """
        zip_file = 'lambda-holder.zip'
        with zipfile.ZipFile(zip_file, mode='w') as zipped:
            zipped.writestr('index.py', 'print "Hello world"')

        contents = ''
        with open('lambda-holder.zip', 'rb') as openfile:
            contents = openfile.read()

        LOG.info('Creating lambda function: %s', self.app_name)

        default_tags = {'app_group': self.group, 'app_name': self.app_name}
        lambda_tags = {**default_tags, **self.custom_tags}

        try:
            self.lambda_client.create_function(
                Environment=self.lambda_environment,
                FunctionName=self.app_name,
                Runtime=self.runtime,
                Role=self.role_arn,
                Handler=self.handler,
                Code={'ZipFile': contents},
                Description=self.description,
                Timeout=int(self.timeout),
                MemorySize=int(self.memory),
                Publish=False,
                VpcConfig=vpc_config,
                Tags=lambda_tags,
                Layers=self.lambda_layers,
                DeadLetterConfig=self.lambda_dlq,
                TracingConfig=self.lambda_tracing,
                FileSystemConfigs=self.lambda_filesystems)

            if self.lambda_destinations:
                self.lambda_client.put_function_event_invoke_config(
                    FunctionName=self.app_name,
                    DestinationConfig=self.lambda_destinations
                )

            if self.lambda_provisioned_throughput:
                self.lambda_client.put_provisioned_concurrency_config(
                    FuctionName=self.app_name,
                    Qualifier=self.env,
                    ProvisionedConcurrentExecutions=self.lambda_provisioned_throughput)

        except boto3.exceptions.botocore.exceptions.ClientError as error:
            if 'CreateNetworkInterface' in error.response['Error']['Message']:
                message = '{0} is missing "ec2:CreateNetworkInterface"'.format(self.role_arn)
                LOG.critical(message)
                raise SystemExit(message)

            raise

        LOG.info("Successfully created Lambda function and alias")

    def create_lambda_function(self):
        """Create or update Lambda function."""
        vpc_config = self._vpc_config()

        if self._check_lambda():
            self.update_function_configuration(vpc_config)
        else:
            self.create_function(vpc_config)

        if self._check_lambda_alias():
            self.update_alias()
        else:
            self.create_alias()



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Retrieve Route 53 Hosted Zone IDs."""
import json
import logging
from pprint import pformat

import boto3
from boto3.exceptions import botocore

from ..consts import DOMAIN
from ..exceptions import PrimaryDNSRecordNotFound
from ..utils import get_template

LOG = logging.getLogger(__name__)


def get_dns_zone_ids(env='dev', facing='internal'):
    """Get Route 53 Hosted Zone IDs for _env_.

    Args:
        env (str): Deployment environment.
        facing (str): Type of ELB, external or internal.

    Returns:
        list: Hosted Zone IDs for _env_. Only *PrivateZone* when _facing_ is
        internal.

    """
    client = boto3.Session(profile_name=env).client('route53')

    zones = client.list_hosted_zones_by_name(DNSName='.'.join([env, DOMAIN]))

    zone_ids = []
    for zone in zones['HostedZones']:
        LOG.debug('Found Hosted Zone: %s', zone)

        if facing == 'external' or zone['Config']['PrivateZone']:
            LOG.info('Using %(Id)s for "%(Name)s", %(Config)s', zone)
            zone_ids.append(zone['Id'])

    LOG.debug('Zone IDs: %s', zone_ids)
    return zone_ids


def update_dns_zone_record(env, zone_id, **kwargs):
    """Create a Route53 CNAME record in _env_ zone.

    Args:
        env (str): Deployment environment.
        zone_id (str): Route53 zone id.

    Keyword Args:
        dns_name (str): FQDN of application's dns entry to add/update.
        dns_name_aws (str): FQDN of AWS resource
        dns_ttl (int): DNS time-to-live (ttl)
    """
    client = boto3.Session(profile_name=env).client('route53')
    response = {}

    hosted_zone_info = client.get_hosted_zone(Id=zone_id)
    zone_name = hosted_zone_info['HostedZone']['Name'].rstrip('.')
    dns_name = kwargs.get('dns_name')

    if dns_name and dns_name.endswith(zone_name):
        dns_name_aws = kwargs.get('dns_name_aws')
        # This is what will be added to DNS
        dns_json = get_template(template_file='infrastructure/dns_upsert.json.j2', **kwargs)
        LOG.info('Attempting to create DNS record %s (%s) in Hosted Zone %s (%s)', dns_name, dns_name_aws, zone_id,
                 zone_name)
        try:
            response = client.change_resource_record_sets(
                HostedZoneId=zone_id,
                ChangeBatch=json.loads(dns_json), )
            LOG.info('Upserted DNS record %s (%s) in Hosted Zone %s (%s)', dns_name, dns_name_aws, zone_id, zone_name)
        except botocore.exceptions.ClientError as error:
            LOG.info('Error creating DNS record %s (%s) in Hosted Zone %s (%s)', dns_name, dns_name_aws, zone_id,
                     zone_name)
            LOG.debug(error)
    else:
        LOG.info('Skipping creating DNS record %s in non-matching Hosted Zone %s (%s)', dns_name, zone_id, zone_name)

    LOG.debug('Route53 JSON Response: \n%s', pformat(response))


def find_existing_record(env, zone_id, dns_name, check_key=None, check_value=None):
    """Check if a specific DNS record exists.

    Args:
        env (str): Deployment environment.
        zone_id (str): Route53 zone id.
        dns_name (str): FQDN of application's dns entry to add/update.
        check_key(str): Key to look for in record. Example: "Type"
        check_value(str): Value to look for with check_key. Example: "CNAME"

    Returns:
        json: Found Record. Returns None if no record found

    """
    client = boto3.Session(profile_name=env).client('route53')
    pager = client.get_paginator('list_resource_record_sets')
    existingrecord = None
    for rset in pager.paginate(HostedZoneId=zone_id):
        for record in rset['ResourceRecordSets']:
            if check_key:
                if record['Name'].rstrip('.') == dns_name and record.get(check_key) == check_value:
                    LOG.info("Found existing record: %s", record)
                    existingrecord = record
                    break
    return existingrecord


def delete_existing_cname(env, zone_id, dns_name):
    """Delete an existing CNAME record.

    This is used when updating to multi-region for deleting old records. The
    record can not just be upserted since it changes types.

    Args:
        env (str): Deployment environment.
        zone_id (str): Route53 zone id.
        dns_name (str): FQDN of application's dns entry to add/update.
    """
    client = boto3.Session(profile_name=env).client('route53')
    startrecord = None
    newrecord_name = dns_name
    startrecord = find_existing_record(env, zone_id, newrecord_name, check_key='Type', check_value='CNAME')
    if startrecord:
        LOG.info("Deleting old record: %s", newrecord_name)
        _response = client.change_resource_record_sets(
            HostedZoneId=zone_id, ChangeBatch={'Changes': [{
                'Action': 'DELETE',
                'ResourceRecordSet': startrecord
            }]})
        LOG.debug('Response from deleting %s: %s', dns_name, _response)


def update_failover_dns_record(env, zone_id, **kwargs):
    """Create a Failover Route53 alias record in _env_ zone.

    Args:
        env (str): Deployment environment.
        zone_id (str): Route53 zone id.

    Keyword Args:
        dns_name (str): FQDN of application's dns entry to add/update.
        dns_ttl (int): DNS time-to-live (ttl)
        elb_aws_dns (str): DNS A Record of ELB from AWS
        elb_dns_zone_id (str): Zone ID of ELB DNS
        failover_state (str): if the record is primary or secondary
        primary_region (str): Primary AWS region for DNS
    """
    client = boto3.Session(profile_name=env).client('route53')
    response = {}

    hosted_zone_info = client.get_hosted_zone(Id=zone_id)
    zone_name = hosted_zone_info['HostedZone']['Name'].rstrip('.')
    dns_name = kwargs.get('dns_name')

    # Check that the primary record exists
    failover_state = kwargs.get('failover_state')
    if failover_state.lower() != 'primary':
        primary_record = find_existing_record(env, zone_id, dns_name, check_key='Failover', check_value='PRIMARY')
        if not primary_record:
            raise PrimaryDNSRecordNotFound("Primary Failover DNS record not found: {}".format(dns_name))

    if dns_name and dns_name.endswith(zone_name):
        dns_json = get_template(template_file='infrastructure/dns_failover_upsert.json.j2', **kwargs)
        LOG.info('Attempting to create DNS Failover record %s (%s) in Hosted Zone %s (%s)', dns_name,
                 kwargs['elb_aws_dns'], zone_id, zone_name)
        try:
            delete_existing_cname(env, zone_id, dns_name)
            response = client.change_resource_record_sets(
                HostedZoneId=zone_id,
                ChangeBatch=json.loads(dns_json), )
            LOG.info('Upserted DNS Failover record %s (%s) in Hosted Zone %s (%s)', dns_name, kwargs['elb_aws_dns'],
                     zone_id, zone_name)
        except botocore.exceptions.ClientError as error:
            LOG.info('Error creating DNS Failover record %s (%s) in Hosted Zone %s (%s)', dns_name,
                     kwargs['elb_aws_dns'], zone_id, zone_name)
            LOG.debug(error)
    else:
        LOG.info('Skipping creating DNS record %s in non-matching Hosted Zone %s (%s)', dns_name, zone_id, zone_name)

    LOG.debug('Route53 JSON Response: \n%s', pformat(response))



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Lookup AMI ID from a simple name."""
import json
import logging
import os
from base64 import b64decode

import gitlab
import requests

from ..consts import AMI_JSON_URL, GIT_URL, GITLAB_TOKEN
from ..exceptions import GitLabApiError
from .warn_user import warn_user

LOG = logging.getLogger(__name__)


def ami_lookup(region='us-east-1', name='tomcat8'):
    """Look up AMI ID.

    Use _name_ to find AMI ID. If no ami_base_url or gitlab_token is provided,
    _name_ is returned as the ami id.

    Args:
        region (str): AWS Region to find AMI ID.
        name (str): Simple AMI base name to lookup.

    Returns:
        str: AMI ID for _name_ in _region_.

    """
    if AMI_JSON_URL:
        ami_dict = _get_ami_dict(AMI_JSON_URL)
        ami_id = ami_dict[region][name]
    elif GITLAB_TOKEN:
        warn_user('Use AMI_JSON_URL feature instead.')
        ami_contents = _get_ami_file(region=region)
        ami_dict = json.loads(ami_contents)
        ami_id = ami_dict[name]
    else:
        ami_id = name

    LOG.info('Using AMI: %s', ami_id)

    return ami_id


def _get_ami_file(region='us-east-1'):
    """Get file from Gitlab.

    Args:
        region (str): AWS Region to find AMI ID.

    Returns:
        str: Contents in json format.

    """
    LOG.info("Getting AMI from Gitlab")
    lookup = FileLookup(git_short='devops/ansible')
    filename = 'scripts/{0}.json'.format(region)
    ami_contents = lookup.remote_file(filename=filename, branch='master')
    LOG.debug('AMI file contents in %s: %s', filename, ami_contents)
    return ami_contents


def _get_ami_dict(json_url):
    """Get ami from a web url.

    Args:
        region (str): AWS Region to find AMI ID.

    Returns:
        dict: Contents in dictionary format.

    """
    LOG.info("Getting AMI from %s", json_url)
    response = requests.get(json_url)
    assert response.ok, "Error getting ami info from {}".format(json_url)
    ami_dict = response.json()
    LOG.debug('AMI json contents: %s', ami_dict)
    return ami_dict


class FileLookup():
    """Retrieve files from a local filesystem or remote GitLab Server.

    When _runway_dir_ is specified, the local directory is given priority and
    remote Git Server will not be used.

    Args:
        git_short (str): Short Git representation of repository, e.g.
            forrest/core.
        runway_dir (str): Root of local runway directory to use instead of
            accessing Git.
    """

    def __init__(self, git_short='', runway_dir=''):
        self.git_short = git_short
        self.runway_dir = os.path.expandvars(os.path.expanduser(runway_dir))

        self.server = None
        self.project = None

        if not self.runway_dir:
            self.get_gitlab_project()

    def get_gitlab_project(self):
        """Get numerical GitLab Project ID.

        Returns:
            int: Project ID number.

        Raises:
            foremast.exceptions.GitLabApiError: GitLab responded with bad status
                code.

        """
        self.server = gitlab.Gitlab(GIT_URL, private_token=GITLAB_TOKEN, api_version=4)
        project = self.server.projects.get(self.git_short)

        if not project:
            raise GitLabApiError('Could not get Project "{0}" from GitLab API.'.format(self.git_short))

        self.project = project
        return self.project

    def local_file(self, filename):
        """Read the local file in _self.runway_dir_.

        Args:
            filename (str): Name of file to retrieve relative to root of
                _runway_dir_.

        Returns:
            str: Contents of local file.

        Raises:
            FileNotFoundError: Requested file missing.

        """
        LOG.info('Retrieving "%s" from "%s".', filename, self.runway_dir)

        file_contents = ''

        file_path = os.path.join(self.runway_dir, filename)

        try:
            with open(file_path, 'rt') as lookup_file:
                file_contents = lookup_file.read()
        except FileNotFoundError:
            LOG.warning('File missing "%s".', file_path)
            raise

        LOG.debug('Local file contents:\n%s', file_contents)
        return file_contents

    def remote_file(self, branch='master', filename=''):
        """Read the remote file on Git Server.

        Args:
            branch (str): Git Branch to find file.
            filename (str): Name of file to retrieve relative to root of
                repository.

        Returns:
            str: Contents of remote file.

        Raises:
            FileNotFoundError: Requested file missing.

        """
        LOG.info('Retrieving "%s" from "%s".', filename, self.git_short)

        file_contents = ''

        try:
            file_blob = self.project.files.get(file_path=filename, ref=branch)
        except gitlab.exceptions.GitlabGetError:
            file_blob = None

        LOG.debug('GitLab file response:\n%s', file_blob)

        if not file_blob:
            msg = 'Project "{0}" is missing file "{1}" in "{2}" branch.'.format(self.git_short, filename, branch)
            LOG.warning(msg)
            raise FileNotFoundError(msg)

        file_contents = b64decode(file_blob.content).decode()

        LOG.debug('Remote file contents:\n%s', file_contents)
        return file_contents

    def get(self, branch='master', filename=''):
        """Retrieve _filename_ from GitLab.

        Args:
            branch (str): Git Branch to find file.
            filename (str): Name of file to retrieve relative to root of Git
                repository, or _runway_dir_ if specified.

        Returns:
            str: Contents of file.

        """
        file_contents = ''

        if self.runway_dir:
            file_contents = self.local_file(filename=filename)
        else:
            file_contents = self.remote_file(branch=branch, filename=filename)

        return file_contents

    def json(self, branch='master', filename=''):
        """Retrieve _filename_ from GitLab.

        Args:
            branch (str): Git Branch to find file.
            filename (str): Name of file to retrieve.

        Returns:
            dict: Decoded JSON.

        Raises:
            SystemExit: Invalid JSON provided.

        """
        file_contents = self.get(branch=branch, filename=filename)

        try:
            json_dict = json.loads(file_contents)
        # TODO: Use json.JSONDecodeError when Python 3.4 has been deprecated
        except ValueError as error:
            msg = ('"{filename}" appears to be invalid json. '
                   'Please validate it with http://jsonlint.com. '
                   'JSON decoder error:\n'
                   '{error}').format(
                       filename=filename, error=error)
            raise SystemExit(msg)

        LOG.debug('JSON object:\n%s', json_dict)
        return json_dict



"""Verify AWS Lambda Function creation."""
import copy
from unittest import mock

from foremast.awslambda.awslambda import LambdaFunction

TEST_PROPERTIES = {
    'pipeline': {
        'lambda': {
            'app_description': None,
            'handler': None,
            'runtime': None,
            'vpc_enabled': None,
        },
    },
    'app': {
        'custom_tags': {},
        'lambda_memory': 0,
        'lambda_timeout': 0,        
        'lambda_environment': None,
        'lambda_layers': None,
        'lambda_dlq': None,
        'lambda_tracing': None,
        'lambda_destinations': None,
        'lambda_subnet_count': None,
        'lambda_filesystems': None,
        'lambda_provisioned_throughput': None,
    }
}
GENERATED_IAM = {
    'lambda_role': 'generated_role',
}


@mock.patch('foremast.awslambda.awslambda.boto3')
@mock.patch('foremast.awslambda.awslambda.get_details')
@mock.patch('foremast.awslambda.awslambda.get_properties')
@mock.patch('foremast.awslambda.awslambda.get_role_arn')
def test_role_arn(mock_get_role_arn, mock_get_properties, mock_get_details, mock_boto3):
    """Check Role ARN configuration."""
    generated = copy.deepcopy(GENERATED_IAM)
    properties = copy.deepcopy(TEST_PROPERTIES)

    mock_get_details.return_value.iam.return_value = generated
    mock_get_properties.return_value = properties

    LambdaFunction(app='test_app', env='test_env', region='us-east-1', prop_path='other')
    mock_get_role_arn.assert_called_with(generated['lambda_role'], mock.ANY, mock.ANY)


@mock.patch('foremast.awslambda.awslambda.boto3')
@mock.patch('foremast.awslambda.awslambda.get_details')
@mock.patch('foremast.awslambda.awslambda.get_properties')
@mock.patch('foremast.awslambda.awslambda.get_role_arn')
def test_role_arn_none(mock_get_role_arn, mock_get_properties, mock_get_details, mock_boto3):
    """Generated Role should be used for Lambda."""
    generated = copy.deepcopy(GENERATED_IAM)
    properties = copy.deepcopy(TEST_PROPERTIES)
    properties['app']['lambda_role'] = None

    mock_get_details.return_value.iam.return_value = generated
    mock_get_properties.return_value = properties

    LambdaFunction(app='test_app', env='test_env', region='us-east-1', prop_path='other')
    mock_get_role_arn.assert_called_with(GENERATED_IAM['lambda_role'], mock.ANY, mock.ANY)


@mock.patch('foremast.awslambda.awslambda.boto3')
@mock.patch('foremast.awslambda.awslambda.get_details')
@mock.patch('foremast.awslambda.awslambda.get_properties')
@mock.patch('foremast.awslambda.awslambda.get_role_arn')
def test_role_arn_custom(mock_get_role_arn, mock_get_properties, mock_get_details, mock_boto3):
    """Custom Role should be used for Lambda."""
    custom_role = 'custom_role'
    generated = copy.deepcopy(GENERATED_IAM)
    properties = copy.deepcopy(TEST_PROPERTIES)
    properties['app']['lambda_role'] = custom_role

    mock_get_details.return_value.iam.return_value = generated
    mock_get_properties.return_value = properties

    LambdaFunction(app='test_app', env='test_env', region='us-east-1', prop_path='other')
    mock_get_role_arn.assert_called_with(custom_role, mock.ANY, mock.ANY)



#   Foremast - Pipeline Tooling
#
#   Copyright 2019 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Set of utility functions for Foremast Google IAP Authentication with Spinnakers API"""
import logging

import google.auth
import google.auth.app_engine
import google.auth.compute_engine.credentials
import google.auth.iam
import google.oauth2.credentials
import google.oauth2.service_account
from google.auth.transport.requests import Request as GoogleAuthRequest

GOOGLE_IAP_IAM_SCOPE = 'https://www.googleapis.com/auth/iam'  # Used in request to Google for OIDC Scope
GOOGLE_OAUTH_TOKEN_URI = 'https://www.googleapis.com/oauth2/v4/token'  # Endpoint for getting tokens for Id Aware Proxy
LOG = logging.getLogger(__name__)


def get_google_iap_bearer_token(client_id, key_path):
    """Makes a request to an application protected by Identity-Aware Proxy.

    Args:
      client_id: The OpenID Connect client ID used by Identity-Aware Proxy.
      key_path: Path to Google Cloud Service Account Credentials in JSON Format

    Returns:
      Mapping[str, str]: The JSON-decoded response data.
    """

    LOG.debug("Parsing Google Cloud Service Account credentials")
    # Load the json svc account credentials
    creds_from_json = google.oauth2.service_account.Credentials.from_service_account_file(
        key_path,
        scopes=[GOOGLE_IAP_IAM_SCOPE],
    )

    LOG.debug("Parsed Google Cloud Service Account credentials with email %s", creds_from_json.signer_email)
    # Construct OAuth 2.0 service account credentials using the signer
    # and email acquired from the json creds file
    service_account_credentials = google.oauth2.service_account.Credentials(
        creds_from_json.signer, creds_from_json.signer_email, token_uri=GOOGLE_OAUTH_TOKEN_URI, additional_claims={
            'target_audience': client_id
        })

    # Create jtw signed by svc account.  This is sent to Google to get a Google signed jwt token
    LOG.debug("Getting service account JWT from endpoint %s", GOOGLE_OAUTH_TOKEN_URI)
    service_account_jwt = (
        service_account_credentials._make_authorization_grant_assertion())
    # Get Google Signed JWT Token
    request = GoogleAuthRequest()
    body = {
        'assertion': service_account_jwt,
        'grant_type': google.oauth2._client._JWT_GRANT_TYPE,
    }

    return google.oauth2._client._token_endpoint_request(request, GOOGLE_OAUTH_TOKEN_URI, body)



#   Foremast - Pipeline Tooling
#
#   Copyright 2019 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Centralized Methods interacting with the Spinnaker Gate API."""
import logging

import requests

from ..consts import API_URL, GATE_AUTHENTICATION, GATE_CA_BUNDLE, GATE_CLIENT_CERT
from ..exceptions import GoogleIAPTokenError
from .google_iap import get_google_iap_bearer_token

LOG = logging.getLogger(__name__)
OAUTH_ENABLED = False


def gate_request(method='GET', uri=None, headers={}, data={}, params={}):
    """Make a request to Gate's API via various auth methods

    Args:
        method (str): Method to request Gate API; GET or POST
        uri (str): URI path to gate API
    """
    response = None

    url = '{host}{uri}'.format(host=API_URL, uri=uri)

    if GATE_AUTHENTICATION:
        if 'google_iap' in GATE_AUTHENTICATION:
            iap_response = get_google_iap_bearer_token(GATE_AUTHENTICATION['google_iap']['oauth_client_id'],
                                                       GATE_AUTHENTICATION['google_iap']['sa_credentials_path'])

            if 'id_token' not in iap_response:
                raise GoogleIAPTokenError

            headers['Authorization'] = 'Bearer {}'.format(iap_response['id_token'])
            LOG.info('Successfully set Google IAP Bearer Token in Request.')

        elif 'github' in GATE_AUTHENTICATION:
            github_token = GATE_AUTHENTICATION['github']['token']
            headers['Authorization'] = 'Bearer {}'.format(github_token)
            LOG.info('Successfully set Github Bearer Token in Request.')

    method = method.upper()
    if method == 'GET':
        response = requests.get(url, params=params, headers=headers, verify=GATE_CA_BUNDLE, cert=GATE_CLIENT_CERT)
    elif method == 'POST':
        response = requests.post(url, data=data, headers=headers, verify=GATE_CA_BUNDLE, cert=GATE_CLIENT_CERT)
    elif method == 'DELETE':
        response = requests.delete(url, headers=headers, verify=GATE_CA_BUNDLE, cert=GATE_CLIENT_CERT)
    else:
        raise NotImplementedError

    if response.status_code in ['401', '403', '503']:
        response.raise_for_status()

    LOG.info(response.content)
    return response



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create Lambda event triggers."""

import logging

from ..utils import get_properties, remove_all_lambda_permissions
from .api_gateway_event import APIGateway
from .cloudwatch_event import create_cloudwatch_event
from .cloudwatch_log_event import create_cloudwatch_log_event
from .event_source_mapping import create_event_source_mapping_trigger
from .s3_event import create_s3_event
from .sns_event import create_sns_event

LOG = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class LambdaEvent:
    """Manipulate Lambda events."""

    def __init__(self, app=None, env=None, region=None, prop_path=None):
        """Lambda event object.

        Args:
            app (str): Application name
            env (str): Environment/Account
            region (str): AWS Region
            prop_path (str): Path of environment property file
        """
        self.app_name = app
        self.env = env
        self.region = region
        self.prop_path = prop_path
        self.properties = get_properties(properties_file=prop_path, env=env, region=region)

    def create_lambda_events(self):
        """Create all defined lambda events for an lambda application."""

        # Clean up lambda permissions before creating triggers

        remove_all_lambda_permissions(app_name=self.app_name, env=self.env, region=self.region)

        triggers = self.properties['lambda_triggers']
        LOG.debug('Lambda triggers defined: %s', triggers)

        for trigger in triggers:
            if trigger['type'] == 'api-gateway':
                apigateway = APIGateway(
                    app=self.app_name, env=self.env, region=self.region, rules=trigger, prop_path=self.prop_path)
                apigateway.setup_lambda_api()

            if trigger['type'] == 'cloudwatch-event':
                create_cloudwatch_event(app_name=self.app_name, env=self.env, region=self.region, rules=trigger)

            if trigger['type'] == 'cloudwatch-logs':
                create_cloudwatch_log_event(app_name=self.app_name, env=self.env, region=self.region, rules=trigger)

            if trigger['type'] in ['dynamodb-stream', 'kinesis-stream', 'sqs']:
                create_event_source_mapping_trigger(app_name=self.app_name, env=self.env, region=self.region,
                                                    event_source=trigger['type'], rules=trigger)

            if trigger['type'] == 'sns':
                create_sns_event(app_name=self.app_name, env=self.env, region=self.region, rules=trigger)

        # filter all triggers to isolate s3 triggers so we can operate on the entire group
        s3_triggers = [x for x in triggers if x['type'] == 's3']

        # group triggers by unique target bucket
        bucket_triggers = dict()
        for s3_trigger in s3_triggers:
            bucket = s3_trigger.get('bucket')
            if bucket in bucket_triggers:
                bucket_triggers[bucket].append(s3_trigger)
            else:
                bucket_triggers[bucket] = [s3_trigger]

        # apply relevant triggers to each respective bucket all at once.
        for bucket, triggers in bucket_triggers.items():
            create_s3_event(app_name=self.app_name, env=self.env, region=self.region, bucket=bucket, triggers=triggers)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Destroy Security Group Resources."""
import logging

from ...utils import get_template, get_vpc_id, wait_for_task
from ...utils.gate import gate_request

LOG = logging.getLogger(__name__)


def destroy_sg(app='', env='', region='', **_):
    """Destroy Security Group.

    Args:
        app (str): Spinnaker Application name.
        env (str): Deployment environment.
        region (str): Region name, e.g. us-east-1.

    Returns:
        True upon successful completion.
    """
    vpc = get_vpc_id(account=env, region=region)

    uri = '/securityGroups/{env}/{region}/{app}'.format(env=env, region=region, app=app)
    payload = {'vpcId': vpc}
    security_group = gate_request(uri=uri, params=payload)

    if not security_group:
        LOG.info('Nothing to delete.')
    else:
        LOG.info('Found Security Group in %(region)s: %(name)s', security_group)

        destroy_request = get_template('destroy/destroy_sg.json.j2', app=app, env=env, region=region, vpc=vpc)
        wait_for_task(destroy_request)

    return True



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Clean removed Pipelines."""
import logging

import requests

from ..consts import RUNWAY_BASE_PATH
from ..exceptions import SpinnakerPipelineCreationFailed, SpinnakerPipelineDeletionFailed
from ..utils import check_managed_pipeline, get_all_pipelines, normalize_pipeline_name
from ..utils.gate import gate_request

LOG = logging.getLogger(__name__)


def delete_pipeline(app='', pipeline_name=''):
    """Delete _pipeline_name_ from _app_."""
    safe_pipeline_name = normalize_pipeline_name(name=pipeline_name)

    LOG.warning('Deleting Pipeline: %s', safe_pipeline_name)

    uri = '/pipelines/{app}/{pipeline}'.format(app=app, pipeline=safe_pipeline_name)

    response = gate_request(method='DELETE', uri=uri)

    if not response.ok:
        LOG.debug('Delete response code: %d', response.status_code)
        if response.status_code == requests.status_codes.codes['method_not_allowed']:
            raise SpinnakerPipelineDeletionFailed('Failed to delete "{0}" from "{1}", '
                                                  'possibly invalid Pipeline name.'.format(safe_pipeline_name, app))
        LOG.debug('Pipeline missing, no delete required.')

    LOG.debug('Deleted "%s" Pipeline response:\n%s', safe_pipeline_name, response.text)

    return response.text


def clean_pipelines(app='', settings=None):
    """Delete Pipelines for regions not defined in application.json files.

    For Pipelines named **app_name [region]**, _region_ will need to appear
    in at least one application.json file. All other names are assumed
    unamanaged.

    Args:
        app (str): Application name
        settings (dict): imported configuration settings

    Returns:
        True: Upon successful completion.

    Raises:
        SpinnakerPipelineCreationFailed: Missing application.json file from
        `create-configs`.
    """
    pipelines = get_all_pipelines(app=app)
    envs = settings['pipeline']['env']

    LOG.debug('Find Regions in: %s', envs)

    regions = set()
    for env in envs:
        try:
            regions.update(settings[env]['regions'])
        except KeyError:
            error_msg = 'Missing "{}/application-master-{}.json".'.format(RUNWAY_BASE_PATH, env)
            raise SpinnakerPipelineCreationFailed(error_msg)
    LOG.debug('Regions defined: %s', regions)

    for pipeline in pipelines:
        pipeline_name = pipeline['name']

        try:
            region = check_managed_pipeline(name=pipeline_name, app_name=app)
        except ValueError:
            LOG.info('"%s" is not managed.', pipeline_name)
            continue

        LOG.debug('Check "%s" in defined Regions.', region)

        if region not in regions:
            delete_pipeline(app=app, pipeline_name=pipeline_name)

    return True



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Test IAM Policy construction."""
import json
from unittest import mock

from foremast.iam.create_iam import create_iam_resources
from foremast.utils import get_template

EC2_TEMPLATE_NAME = 'infrastructure/iam/trust/ec2_role.json.j2'
LAMBDA_TEMPLATE_NAME = 'infrastructure/iam/trust/lambda_role.json.j2'


@mock.patch('foremast.iam.create_iam.attach_profile_to_role')
@mock.patch('foremast.iam.create_iam.boto3.session.Session')
@mock.patch('foremast.iam.create_iam.construct_policy')
@mock.patch('foremast.iam.create_iam.get_details')
@mock.patch('foremast.iam.create_iam.get_properties')
@mock.patch('foremast.iam.create_iam.resource_action')
def test_create_iam_resources(resource_action, get_properties, get_details, construct_policy, session,
                              attach_profile_to_role):
    """Check basic functionality."""
    get_details.return_value.iam.return_value = {'group': 1, 'policy': 2, 'profile': 3, 'role': 4, 'user': 5}
    get_properties.return_value = {'type': 'ec2'}

    assert create_iam_resources(env='narnia', app='lion/aslan')

    assert resource_action.call_count == 6
    session.assert_called_with(profile_name='narnia')
    get_details.assert_called_with(env='narnia', app='lion/aslan')
    get_properties.assert_called_with(env='pipeline')
    construct_policy.assert_called_with(
        app='lion/aslan', group=1, env='narnia', pipeline_settings=get_properties.return_value)


@mock.patch('foremast.iam.create_iam.attach_profile_to_role')
@mock.patch('foremast.iam.create_iam.boto3.session.Session')
@mock.patch('foremast.iam.create_iam.construct_policy')
@mock.patch('foremast.iam.create_iam.get_details')
@mock.patch('foremast.iam.create_iam.get_properties')
@mock.patch('foremast.iam.create_iam.get_template')
@mock.patch('foremast.iam.create_iam.resource_action')
def test_iam_role_policy(resource_action, get_template, get_properties, get_details, construct_policy, session,
                         attach_profile_to_role):
    """IAM Role Policy should match deployment type."""
    get_properties.return_value = {'type': 'ec2'}
    get_details.return_value.iam.return_value = {'group': 1, 'policy': 2, 'profile': 3, 'role': 4, 'user': 5}

    assert create_iam_resources()

    get_template.assert_called_with(EC2_TEMPLATE_NAME, formats=get_details())
    calls = [
        mock.call(
            mock.ANY,
            action='create_role',
            log_format=mock.ANY,
            RoleName=mock.ANY,
            AssumeRolePolicyDocument=get_template.return_value)
    ]
    resource_action.assert_has_calls(calls)


def test_ec2_iam_policy():
    """Check template for proper format."""
    ec2_json = get_template(EC2_TEMPLATE_NAME)
    ec2_dict = json.loads(ec2_json)

    assert all(key in ec2_dict for key in ('Version', 'Statement'))
    assert len(ec2_dict['Statement']) == 1

    only_statement = ec2_dict['Statement'][0]

    assert only_statement['Action'] == 'sts:AssumeRole'
    assert only_statement['Effect'] == 'Allow'
    assert only_statement['Principal']['Service'] == 'ec2.amazonaws.com'


def test_lambda_iam_policy():
    """Check Lambda Trust Relationship template format."""
    lambda_json = get_template(LAMBDA_TEMPLATE_NAME)
    lambda_dict = json.loads(lambda_json)

    assert all(key in lambda_dict for key in ('Version', 'Statement'))
    assert len(lambda_dict['Statement']) == 1

    only_statement = lambda_dict['Statement'][0]

    assert only_statement['Action'] == 'sts:AssumeRole'
    assert only_statement['Effect'] == 'Allow'
    assert only_statement['Principal']['Service'] == 'lambda.amazonaws.com'



"""SNS Subscription functions."""
import logging

import boto3

from ..utils.awslambda import get_lambda_alias_arn

LOG = logging.getLogger(__name__)


def get_sns_subscriptions(app_name, env, region):
    """List SNS lambda subscriptions.

    Returns:
        list: List of Lambda subscribed SNS ARNs.

    """
    session = boto3.Session(profile_name=env, region_name=region)
    sns_client = session.client('sns')

    lambda_alias_arn = get_lambda_alias_arn(app=app_name, account=env, region=region)

    lambda_subscriptions = []
    subscriptions = sns_client.list_subscriptions()

    for subscription in subscriptions['Subscriptions']:
        if subscription['Protocol'] == "lambda" and subscription['Endpoint'] == lambda_alias_arn:
            lambda_subscriptions.append(subscription['SubscriptionArn'])

    if not lambda_subscriptions:
        LOG.debug('SNS subscription for function %s not found', lambda_alias_arn)

    return lambda_subscriptions



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create Pipelines for Spinnaker."""
import collections
import json
from pprint import pformat

from ..consts import DEFAULT_RUN_AS_USER
from ..utils import get_template
from .clean_pipelines import clean_pipelines
from .construct_pipeline_block_datapipeline import construct_datapipeline
from .create_pipeline import SpinnakerPipeline
from .renumerate_stages import renumerate_stages


class SpinnakerPipelineDataPipeline(SpinnakerPipeline):
    """Manipulate Spinnaker Pipelines.

    Args:
        app (str): Application name.
        trigger_job (str): Jenkins trigger job.
        base (str): Base image name (i.e: fedora).
        prop_path (str): Path to the raw.properties.json.
    """

    def render_wrapper(self, region='us-east-1'):
        """Generate the base Pipeline wrapper.

        This renders the non-repeatable stages in a pipeline, like jenkins, baking, tagging and notifications.

        Args:
            region (str): AWS Region.

        Returns:
            dict: Rendered Pipeline wrapper.
        """
        base = self.base or self.settings['pipeline']['base']

        email = self.settings['pipeline']['notifications']['email']
        slack = self.settings['pipeline']['notifications']['slack']
        deploy_type = self.settings['pipeline']['type']
        pipeline_id = self.compare_with_existing(region=region)

        data = {
            'app': {
                'appname': self.app_name,
                'group_name': self.group_name,
                'repo_name': self.repo_name,
                'base': base,
                'deploy_type': deploy_type,
                'environment': 'packaging',
                'region': region,
                'triggerjob': self.trigger_job,
                'run_as_user': DEFAULT_RUN_AS_USER,
                'email': email,
                'slack': slack,
                'pipeline': self.settings['pipeline']
            },
            'id': pipeline_id
        }

        self.log.debug('Wrapper app data:\n%s', pformat(data))

        wrapper = get_template(template_file='pipeline/pipeline_wrapper.json.j2', data=data, formats=self.generated)

        return json.loads(wrapper)

    def create_pipeline(self):
        """Main wrapper for pipeline creation.
        1. Runs clean_pipelines to clean up existing ones
        2. determines which environments the pipeline needs
        3. Renders all of the pipeline blocks as defined in configs
        4. Runs post_pipeline to create pipeline
        """
        clean_pipelines(app=self.app_name, settings=self.settings)

        pipeline_envs = self.environments
        self.log.debug('Envs from pipeline.json: %s', pipeline_envs)

        regions_envs = collections.defaultdict(list)
        for env in pipeline_envs:
            for region in self.settings[env]['regions']:
                regions_envs[region].append(env)
        self.log.info('Environments and Regions for Pipelines:\n%s', json.dumps(regions_envs, indent=4))

        pipelines = {}
        for region, envs in regions_envs.items():
            # TODO: Overrides for an environment no longer makes sense. Need to
            # provide override for entire Region possibly.
            pipelines[region] = self.render_wrapper(region=region)

            previous_env = None
            for env in envs:

                block = construct_datapipeline(
                    env=env,
                    generated=self.generated,
                    previous_env=previous_env,
                    region=region,
                    settings=self.settings[env][region],
                    pipeline_data=self.settings['pipeline'])
                pipelines[region]['stages'].extend(json.loads(block))

                previous_env = env

        self.log.debug('Assembled Pipelines:\n%s', pformat(pipelines))

        for region, pipeline in pipelines.items():
            renumerate_stages(pipeline)

            self.post_pipeline(pipeline)

        return True



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Construct an IAM Policy from templates.

Examples:
    pipeline.json::

        {
            "services": {
                "dynamodb": [
                    "another_app"
                ],
                "lambda": true,
                "s3": true
            }
        }
"""
import json
import logging

from ..utils import get_env_credential, get_template

LOG = logging.getLogger(__name__)


def auto_service(pipeline_settings={}, services={}):
    """Automatically enable service for deployment types.

    Args:
        services (dict): Services to enable in IAM Policy.
        pipeline_settings (dict): Settings from *pipeline.json*.

    Returns:
        dict: Services.
    """
    deployment_type = pipeline_settings['type']

    if deployment_type == 'lambda':
        services['lambda'] = True

    return services


def render_policy_template(  # pylint: disable=too-many-arguments
        account_number='',
        app='coreforrest',
        env='dev',
        group='forrest',
        items=None,
        pipeline_settings=None,
        region='us-east-1',
        service=''):
    """Render IAM Policy template.

    To support multiple statement blocks, JSON objects can be separated by a
    comma. This function attempts to turn any invalid JSON into a valid list
    based on this comma separated assumption.

    Args:
        account_number (str): AWS Account number.
        app (str): Name of Spinnaker Application.
        env (str): Environment/Account in AWS
        group (str):A Application group/namespace
        items (list): Resource names used to create a Policy per Resource.
        region (str): AWS region.
        pipeline_settings (dict): Settings from *pipeline.json*.
        service (str): Name of cloud service to find matching IAM Policy
            template.

    Returns:
        list: IAM Policy :obj:`dict` statements for the given service.

    """
    statements = []

    rendered_service_policy = get_template(
        'infrastructure/iam/{0}.json.j2'.format(service),
        account_number=account_number,
        app=app,
        env=env,
        group=group,
        region=region,
        items=items,
        settings=pipeline_settings)

    try:
        statement_block = json.loads(rendered_service_policy)
        statements.append(statement_block)
    except ValueError:
        LOG.debug('Need to make %s template into list.', service)
        statements = json.loads('[{0}]'.format(rendered_service_policy))

    LOG.debug('Rendered IAM Policy statements: %s', statements)

    return statements


def construct_policy(app='coreforrest', env='dev', group='forrest', region='us-east-1', pipeline_settings=None):
    """Assemble IAM Policy for _app_.

    Args:
        app (str): Name of Spinnaker Application.
        env (str): Environment/Account in AWS
        group (str):A Application group/namespace
        region (str): AWS region
        pipeline_settings (dict): Settings from *pipeline.json*.

    Returns:
        json: Custom IAM Policy for _app_.
        None: When no *services* have been defined in *pipeline.json*.
    """
    LOG.info('Create custom IAM Policy for %s.', app)

    services = pipeline_settings.get('services', {})
    LOG.debug('Found requested services: %s', services)

    services = auto_service(pipeline_settings=pipeline_settings, services=services)

    if services:
        credential = get_env_credential(env=env)
        account_number = credential['accountId']

    statements = []
    for service, value in services.items():
        if value is True:
            items = []
        elif isinstance(value, str):
            items = [value]
        else:
            items = value

        rendered_statements = render_policy_template(
            account_number=account_number,
            app=app,
            env=env,
            group=group,
            items=items,
            pipeline_settings=pipeline_settings,
            region=region,
            service=service)

        statements.extend(rendered_statements)

    if statements:
        policy_json = get_template('infrastructure/iam/wrapper.json.j2', statements=json.dumps(statements))
    else:
        LOG.info('No services defined for %s.', app)
        policy_json = None

    return policy_json



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Prints a banner.

Example:

================================================================================
                             Create Security Group
================================================================================
"""
import logging

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


def banner(text, border='=', width=80):
    """Center _text_ in a banner _width_ wide with _border_ characters.

    Args:
        text (str): What to write in the banner
        border (str): Border character
        width (int): How long the border should be
    """
    text_padding = '{0:^%d}' % (width)
    LOG.info(border * width)
    LOG.info(text_padding.format(text))
    LOG.info(border * width)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Manages AWS scaling policies in Spinnaker. Can find, create, and delete.

This module also creates an inverse policy for scaling down
"""
import json
import logging
import os
from math import floor

from ..utils import get_latest_server_group, get_properties, get_template, wait_for_task
from ..utils.gate import gate_request


class AutoScalingPolicy:
    """Manages scaling policies in Spinnaker

    Args:
        app (str): Application name
        prop_path (str): Path of rendered property files
        env (str): Environment/Account to add policy to
        region (str): AWS region for policy

    Attributes:
        log (str): Logger name
        settings (dict): Properties imported from prop_path
    """

    def __init__(self, app='', prop_path='', env='', region=''):

        self.log = logging.getLogger(__name__)

        self.here = os.path.dirname(os.path.realpath(__file__))
        self.env = env
        self.region = region
        self.app = app

        self.settings = get_properties(properties_file=prop_path, env=self.env, region=self.region)

    def prepare_policy_template(self, scaling_type, server_group, scaling_policy=None):
        """Renders scaling policy templates based on configs and variables.
        After rendering, POSTs the json to Spinnaker for creation.

        Args:
            scaling_type (str): Type of policy: ``scale_up``, ``scale_down``, ``custom``
            server_group (str): Server group to render and apply policy template to
            scaling_policy (dict): Custom Scaling Policy dictionary, defaults to None.
        """
        if 'period_minutes' in self.settings['asg']['scaling_policy']:
            period_sec = int(self.settings['asg']['scaling_policy']['period_minutes']) * 60
        else:
            period_sec = 1800

        template_kwargs = {
            'app': self.app,
            'env': self.env,
            'region': self.region,
            'server_group': server_group,
            'period_sec': period_sec,
            'scaling_policy': self.settings['asg']['scaling_policy'],
        }

        if scaling_type == 'scale_up':
            scale_up_adjustment = int(self.settings['asg']['scaling_policy'].get('increase_scaling_adjustment', 1))
            template_kwargs['operation'] = 'increase'
            template_kwargs['comparisonOperator'] = 'GreaterThanThreshold'
            template_kwargs['scalingAdjustment'] = scale_up_adjustment
            rendered_template = get_template(template_file='infrastructure/autoscaling_policy.json.j2',
                                             **template_kwargs)
        elif scaling_type == 'scale_down':
            scale_down_adjustment = int(self.settings['asg']['scaling_policy'].get('decrease_scaling_adjustment', -1))
            cur_threshold = int(self.settings['asg']['scaling_policy']['threshold'])
            self.settings['asg']['scaling_policy']['threshold'] = floor(cur_threshold * 0.5)
            template_kwargs['operation'] = 'decrease'
            template_kwargs['comparisonOperator'] = 'LessThanThreshold'
            template_kwargs['scalingAdjustment'] = scale_down_adjustment
            rendered_template = get_template(template_file='infrastructure/autoscaling_policy.json.j2',
                                             **template_kwargs)

        elif scaling_type == 'custom':
            template_kwargs['scaling_policy'] = scaling_policy

            # Helper to update Cluster Name with latest server group name
            for each_dimension in template_kwargs['scaling_policy']['scaling_metric']['dimensions']:
                pos = template_kwargs['scaling_policy']['scaling_metric']['dimensions'].index(each_dimension)
                if each_dimension['name'] == 'AutoScalingGroupName' and each_dimension['value'] == '$self':
                    template_kwargs['scaling_policy']['scaling_metric']['dimensions'][pos]['value'] = server_group

            if scaling_policy['scaling_type'] == 'step_scaling':
                rendered_template = get_template(
                    template_file='infrastructure/autoscaling_custom_stepscaling_policy.json.j2',
                    **template_kwargs)
            elif scaling_policy['scaling_type'] == 'target_tracking':
                rendered_template = get_template(
                    template_file='infrastructure/autoscaling_custom_targettracking_policy.json.j2',
                    **template_kwargs)
            else:
                self.log.warn('Scaling Type %s not implemented or does not exist.', scaling_policy['scaling_type'])
                raise NotImplementedError

        self.log.info('Creating a %s policy in %s for %s', scaling_type, self.env, self.app)
        wait_for_task(rendered_template)
        self.log.info('Successfully created a %s policy in %s for %s', scaling_type, self.env, self.app)

    def create_policy(self):
        """Wrapper function. Gets the server group, sets sane defaults,
        deletes existing policies, and then runs self.prepare_policy_template
        for scaling up and scaling down policies.
        This function acts as the main driver for the scaling policy creationprocess
        """
        if not self.settings['asg']['scaling_policy'] and not self.settings['asg']['custom_scaling_policies']:
            self.log.info("No scaling policy found, skipping...")
            return

        server_group = get_latest_server_group(self.env, self.app)

        # Find all existing and remove them
        scaling_policies = self.get_all_scaling_policies(server_group)
        for policy_block in scaling_policies:
            for scaling_policy in policy_block:
                self.delete_existing_scaling_policy(scaling_policy, server_group)

        if self.settings['asg']['scaling_policy']:
            self.prepare_policy_template('scale_up', server_group)
            if self.settings['asg']['scaling_policy'].get('scale_down', True):
                self.prepare_policy_template('scale_down', server_group)
        elif self.settings['asg']['custom_scaling_policies']:
            for scaling_policy in self.settings['asg']['custom_scaling_policies']:
                self.prepare_policy_template('custom', server_group, scaling_policy)

    def delete_existing_scaling_policy(self, scaling_policy, server_group):
        """Given a scaling_policy and server_group, deletes the existing scaling_policy.
        Scaling policies need to be deleted instead of upserted for consistency.

        Args:
            scaling_policy (json): the scaling_policy json from Spinnaker that should be deleted
            server_group (str): the affected server_group
        """
        self.log.info("Deleting policy %s on %s", scaling_policy['policyName'], server_group)
        delete_dict = {
            "application":
            self.app,
            "description":
            "Delete scaling policy",
            "job": [{
                "policyName": scaling_policy['policyName'],
                "serverGroupName": server_group,
                "credentials": self.env,
                "region": self.region,
                "provider": "aws",
                "type": "deleteScalingPolicy",
                "user": "foremast-autoscaling-policy"
            }]
        }
        wait_for_task(json.dumps(delete_dict))

    def get_all_scaling_policies(self, server_group):
        """Finds all existing scaling policies for an application

        Returns:
            scalingpolicies (list): List of all existing scaling policies for the application
        """
        self.log.info("Checking for existing scaling policy")
        uri = "/applications/{0}/clusters/{1}/{0}/serverGroups".format(self.app, self.env)
        response = gate_request(uri=uri)
        assert response.ok, "Error looking for existing Autoscaling Policy for {0}: {1}".format(self.app, response.text)

        scalingpolicies = []
        for servergroup in response.json():
            if servergroup['scalingPolicies'] and servergroup['asg']['autoScalingGroupName'] == server_group:
                self.log.info("Found policies on %s", server_group)
                scalingpolicies.append(servergroup['scalingPolicies'])
        self.log.debug("Scaling policies: %s", scalingpolicies)
        return scalingpolicies



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Construct a block section of Stages in a Spinnaker Pipeline."""
import copy
import logging
from pprint import pformat

from ..consts import ENV_CONFIGS
from ..utils import get_template, verify_approval_skip

LOG = logging.getLogger(__name__)


def construct_datapipeline(env='',
                           generated=None,
                           previous_env=None,
                           region='us-east-1',
                           settings=None,
                           pipeline_data=None):
    """Create the Pipeline JSON from template.

    This handles the common repeatable patterns in a pipeline, such as
    judgement, infrastructure, tagger and qe.

    Args:
        env (str): Deploy environment name, e.g. dev, stage, prod.
        generated (gogoutils.Generator): Gogo Application name generator.
        previous_env (str): The previous deploy environment to use as
            Trigger.
        region (str): AWS Region to deploy to.
        settings (dict): Environment settings from configurations.

    Returns:
        dict: Pipeline JSON template rendered with configurations.
    """
    LOG.info('%s block for [%s].', env, region)

    if env.startswith('prod'):
        template_name = 'pipeline/pipeline_{}_datapipeline.json.j2'.format(env)
    else:
        template_name = 'pipeline/pipeline_stages_datapipeline.json.j2'

    LOG.debug('%s info:\n%s', env, pformat(settings))

    gen_app_name = generated.app_name()

    data = copy.deepcopy(settings)

    approval_skip = verify_approval_skip(data, env, ENV_CONFIGS)

    data['app'].update({
        'appname': gen_app_name,
        'approval_skip': approval_skip,
        'repo_name': generated.repo,
        'group_name': generated.project,
        'environment': env,
        'region': region,
        'previous_env': previous_env,
        'promote_restrict': pipeline_data['promote_restrict'],
        'owner_email': pipeline_data['owner_email']
    })

    LOG.debug('Block data:\n%s', pformat(data))

    pipeline_json = get_template(template_file=template_name, data=data, formats=generated)
    return pipeline_json



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Lambda related utilities."""
import json
import logging

import boto3

from ..exceptions import LambdaAliasDoesNotExist, LambdaFunctionDoesNotExist

LOG = logging.getLogger(__name__)
FOREMAST_PREFIX = "foremast-"


def get_lambda_arn(app, account, region):
    """Get lambda ARN.

    Args:
        account (str): AWS account name.
        region (str): Region name, e.g. us-east-1
        app (str): Lambda function name

    Returns:
        str: ARN for requested lambda function

    """
    session = boto3.Session(profile_name=account, region_name=region)
    lambda_client = session.client('lambda')

    lambda_arn = None
    paginator = lambda_client.get_paginator('list_functions')

    for lambda_functions in paginator.paginate():
        for lambda_function in lambda_functions['Functions']:
            if lambda_function['FunctionName'] == app:
                lambda_arn = lambda_function['FunctionArn']
                LOG.debug("Lambda ARN for lambda function %s is %s.", app, lambda_arn)
                break
        if lambda_arn:
            break

    if not lambda_arn:
        LOG.fatal('Lambda function with name %s not found in %s %s', app, account, region)
        raise LambdaFunctionDoesNotExist(
            'Lambda function with name {0} not found in {1} {2}'.format(app, account, region))

    return lambda_arn


def get_lambda_alias_arn(app, account, region):
    """Get lambda alias ARN. Assumes that account name is equal to alias name.

    Args:
        account (str): AWS account name.
        region (str): Region name, e.g. us-east-1
        app (str): Lambda function name

    Returns:
        str: ARN for requested lambda alias

    """
    session = boto3.Session(profile_name=account, region_name=region)
    lambda_client = session.client('lambda')

    lambda_aliases = lambda_client.list_aliases(FunctionName=app)

    matched_alias = None
    for alias in lambda_aliases['Aliases']:
        if alias['Name'] == account:
            lambda_alias_arn = alias['AliasArn']
            LOG.info('Found ARN for alias %s for function %s', account, app)
            matched_alias = lambda_alias_arn
            break
    else:
        fatal_message = 'Lambda alias {0} of function {1} not found'.format(account, app)
        LOG.fatal(fatal_message)
        raise LambdaAliasDoesNotExist(fatal_message)
    return matched_alias


def add_lambda_permissions(function='',
                           statement_id='',
                           action='lambda:InvokeFunction',
                           principal='',
                           source_arn='',
                           env='',
                           region='us-east-1'):
    """Add permission to Lambda for the event trigger.

    Args:
        function (str): Lambda function name
        statement_id (str): IAM policy statement (principal) id
        action (str): Lambda action to allow
        principal (str): AWS principal to add permissions
        source_arn (str): ARN of the source of the event.
        env (str): Environment/account of function
        region (str): AWS region of function
    """
    session = boto3.Session(profile_name=env, region_name=region)
    lambda_client = session.client('lambda')
    response_action = None
    prefixed_sid = FOREMAST_PREFIX + statement_id

    add_permissions_kwargs = {
        'FunctionName': function,
        'StatementId': prefixed_sid,
        'Action': action,
        'Principal': principal,
    }

    if source_arn:
        add_permissions_kwargs['SourceArn'] = source_arn

    try:
        lambda_client.add_permission(**add_permissions_kwargs)
        response_action = 'Add permission with Sid: {}'.format(prefixed_sid)
    except boto3.exceptions.botocore.exceptions.ClientError as error:
        LOG.debug('Add permission error: %s', error)
        response_action = "Did not add permissions"

    LOG.debug('Related StatementId (SID): %s', prefixed_sid)
    LOG.info(response_action)


def remove_all_lambda_permissions(app_name='', env='', region='us-east-1'):
    """Remove all foremast-* permissions from lambda.

    Args:
        app_name (str): Application name
        env (str): AWS environment
        region (str): AWS region
    """
    session = boto3.Session(profile_name=env, region_name=region)
    lambda_client = session.client('lambda')
    legacy_prefix = app_name + "_"

    lambda_arn = get_lambda_arn(app_name, env, region)
    lambda_alias_arn = get_lambda_alias_arn(app_name, env, region)
    arns = (lambda_arn, lambda_alias_arn)

    for arn in arns:
        try:
            response = lambda_client.get_policy(FunctionName=arn)
        except boto3.exceptions.botocore.exceptions.ClientError as error:
            LOG.info("No policy exists for function %s, skipping deletion", arn)
            LOG.debug(error)
            continue

        policy_json = json.loads(response['Policy'])
        LOG.debug("Found Policy: %s", response)
        for perm in policy_json['Statement']:
            if perm['Sid'].startswith(FOREMAST_PREFIX) or perm['Sid'].startswith(legacy_prefix):
                lambda_client.remove_permission(FunctionName=arn, StatementId=perm['Sid'])
                LOG.info('removed permission: %s', perm['Sid'])
            else:
                LOG.info('Skipping deleting permission %s - Not managed by Foremast', perm['Sid'])



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Add the appropriate ELB Listeners."""
import json
import logging

from ..exceptions import ForemastTemplateNotFound
from ..utils import get_env_credential, get_template

LOG = logging.getLogger(__name__)


def format_listeners(elb_settings=None, env='dev', region='us-east-1'):
    """Format ELB Listeners into standard list.

    Args:
        elb_settings (dict): ELB settings including ELB Listeners to add,
            e.g.::

                # old
                {
                    "certificate": null,
                    "i_port": 8080,
                    "lb_port": 80,
                    "subnet_purpose": "internal",
                    "target": "HTTP:8080/health"
                }

                # new
                {
                    "ports": [
                        {
                            "instance": "HTTP:8080",
                            "loadbalancer": "HTTP:80"
                        },
                        {
                            "certificate": "cert_name",
                            "instance": "HTTP:8443",
                            "loadbalancer": "HTTPS:443"
                        }
                    ],
                    "subnet_purpose": "internal",
                    "target": "HTTP:8080/health"
                }

        env (str): Environment to find the Account Number for.

    Returns:
        list: ELB Listeners formatted into dicts for Spinnaker::

            [
                {
                    'externalPort': 80,
                    'externalProtocol': 'HTTP',
                    'internalPort': 8080,
                    'internalProtocol': 'HTTP',
                    'sslCertificateId': None,
                    'listenerPolicies': [],
                    'backendPolicies': []
                },
                ...
            ]
    """
    LOG.debug('ELB settings:\n%s', elb_settings)

    credential = get_env_credential(env=env)
    account = credential['accountId']

    listeners = []

    if 'ports' in elb_settings:
        for listener in elb_settings['ports']:
            cert_name = format_cert_name(
                env=env, region=region, account=account, certificate=listener.get('certificate', None))

            lb_proto, lb_port = listener['loadbalancer'].split(':')
            i_proto, i_port = listener['instance'].split(':')
            listener_policies = listener.get('policies', [])
            listener_policies += listener.get('listener_policies', [])
            backend_policies = listener.get('backend_policies', [])

            elb_data = {
                'externalPort': int(lb_port),
                'externalProtocol': lb_proto.upper(),
                'internalPort': int(i_port),
                'internalProtocol': i_proto.upper(),
                'sslCertificateId': cert_name,
                'listenerPolicies': listener_policies,
                'backendPolicies': backend_policies,
            }

            listeners.append(elb_data)
    else:
        listener_policies = elb_settings.get('policies', [])
        listener_policies += elb_settings.get('listener_policies', [])
        backend_policies = elb_settings.get('backend_policies', [])

        listeners = [{
            'externalPort': int(elb_settings['lb_port']),
            'externalProtocol': elb_settings['lb_proto'],
            'internalPort': int(elb_settings['i_port']),
            'internalProtocol': elb_settings['i_proto'],
            'sslCertificateId': elb_settings['certificate'],
            'listenerPolicies': listener_policies,
            'backendPolicies': backend_policies,
        }]

    for listener in listeners:
        LOG.info('ELB Listener:\n'
                 'loadbalancer %(externalProtocol)s:%(externalPort)d\n'
                 'instance %(internalProtocol)s:%(internalPort)d\n'
                 'certificate: %(sslCertificateId)s\n'
                 'listener_policies: %(listenerPolicies)s\n'
                 'backend_policies: %(backendPolicies)s', listener)
    return listeners


def format_cert_name(env='', account='', region='', certificate=None):
    """Format the SSL certificate name into ARN for ELB.

    Args:
        env (str): Account environment name
        account (str): Account number for ARN
        region (str): AWS Region.
        certificate (str): Name of SSL certificate

    Returns:
        str: Fully qualified ARN for SSL certificate
        None: Certificate is not desired
    """
    cert_name = None

    if certificate:
        if certificate.startswith('arn'):
            LOG.info("Full ARN provided...skipping lookup.")
            cert_name = certificate
        else:
            generated_cert_name = generate_custom_cert_name(env, region, account, certificate)
            if generated_cert_name:
                LOG.info("Found generated certificate %s from template", generated_cert_name)
                cert_name = generated_cert_name
            else:
                LOG.info("Using default certificate name logic")
                cert_name = ('arn:aws:iam::{account}:server-certificate/{name}'.format(
                    account=account, name=certificate))
    LOG.debug('Certificate name: %s', cert_name)

    return cert_name


def generate_custom_cert_name(env='', region='', account='', certificate=None):
    """Generate a custom TLS Cert name based on a template.

    Args:
        env (str): Account environment name
        region (str): AWS Region.
        account (str): Account number for ARN.
        certificate (str): Name of SSL certificate.

    Returns:
        str: Fully qualified ARN for SSL certificate.
        None: Template doesn't exist.
    """
    cert_name = None
    template_kwargs = {'account': account, 'name': certificate}

    # TODO: Investigate moving this to a remote API, then fallback to local file if unable to connect
    try:
        rendered_template = get_template(template_file='infrastructure/iam/tlscert_naming.json.j2', **template_kwargs)
        tlscert_dict = json.loads(rendered_template)
    except ForemastTemplateNotFound:
        LOG.info('Unable to find TLS Cert Template...falling back to default logic...')
        return cert_name

    # TODO: Move to v1 method for check
    try:
        LOG.info("Attempting to find TLS Cert using TLS Cert Template v1 lookup...")
        cert_name = tlscert_dict[env][certificate]
        LOG.info("Found TLS certificate named %s under %s using TLS Cert Template v1", certificate, env)
    except KeyError:
        LOG.error("Unable to find TLS certificate named %s under %s using v1 TLS Cert Template.", certificate, env)

    # TODO: Move variable to consts
    # TODO: move to v2 method for check
    tls_services = ['iam', 'acm']
    if cert_name is None and all(service in tlscert_dict for service in tls_services):
        LOG.info("Attempting to find TLS Cert using TLS Cert Template v2 lookup...")
        if certificate in tlscert_dict['iam'][env]:
            cert_name = tlscert_dict['iam'][env][certificate]
            LOG.info("Found IAM TLS certificate named %s under %s using TLS Cert Template v2", certificate, env)
        elif certificate in tlscert_dict['acm'][region][env]:
            cert_name = tlscert_dict['acm'][region][env][certificate]
            LOG.info("Found ACM TLS certificate named %s under %s in %s using TLS Cert Template v2", certificate, env,
                     region)
        else:
            LOG.error(
                "Unable to find TLS certificate named %s under parent keys [ACM, IAM] %s in v2 TLS Cert Template.",
                certificate, env)

    return cert_name



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Destroy any DNS records."""
import logging

import boto3

from ....utils.get_cloudwatch_event_rule import get_cloudwatch_event_rule

LOG = logging.getLogger(__name__)


def destroy_cloudwatch_event(app='', env='dev', region=''):
    """Destroy Cloudwatch event subscription.

    Args:
        app (str): Spinnaker Application name.
        env (str): Deployment environment.
        region (str): AWS region.
    Returns:
        bool: True upon successful completion.
    """

    session = boto3.Session(profile_name=env, region_name=region)
    cloudwatch_client = session.client('events')

    event_rules = get_cloudwatch_event_rule(app_name=app, account=env, region=region)

    for rule in event_rules:
        cloudwatch_client.remove_targets(Rule=rule, Ids=[app])

    return True



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Destroy any IAM related resources."""
import collections
import logging

import boto3

from ...utils import get_details
from ..resource_action import resource_action

LOG = logging.getLogger(__name__)


def destroy_iam(app='', env='dev', **_):
    """Destroy IAM Resources.

    Args:
        app (str): Spinnaker Application name.
        env (str): Deployment environment, i.e. dev, stage, prod.

    Returns:
        True upon successful completion.
    """
    session = boto3.Session(profile_name=env)
    client = session.client('iam')

    generated = get_details(env=env, app=app)
    generated_iam = generated.iam()
    app_details = collections.namedtuple('AppDetails', generated_iam.keys())
    details = app_details(**generated_iam)

    LOG.debug('Application details: %s', details)

    resource_action(
        client,
        action='remove_user_from_group',
        log_format='Removed user from group: %(UserName)s ~> %(GroupName)s',
        GroupName=details.group,
        UserName=details.user)
    resource_action(client, action='delete_user', log_format='Destroyed user: %(UserName)s', UserName=details.user)
    resource_action(client, action='delete_group', log_format='Destroyed group: %(GroupName)s', GroupName=details.group)

    resource_action(
        client,
        action='remove_role_from_instance_profile',
        log_format='Destroyed Instance Profile from Role: '
        '%(InstanceProfileName)s ~> %(RoleName)s',
        InstanceProfileName=details.profile,
        RoleName=details.role)
    resource_action(
        client,
        action='delete_instance_profile',
        log_format='Destroyed Instance Profile: %(InstanceProfileName)s',
        InstanceProfileName=details.profile)

    role_policies = []
    try:
        role_policies = resource_action(
            client,
            action='list_role_policies',
            log_format='Found Role Policies for %(RoleName)s.',
            RoleName=details.role)['PolicyNames']
    except TypeError:
        LOG.info('Role %s not found.', details.role)

    for policy in role_policies:
        resource_action(
            client,
            action='delete_role_policy',
            log_format='Removed Inline Policy from Role: '
            '%(PolicyName)s ~> %(RoleName)s',
            RoleName=details.role,
            PolicyName=policy)

    attached_role_policies = []
    try:
        attached_role_policies = resource_action(
            client,
            action='list_attached_role_policies',
            log_format='Found attached Role Polices for %(RoleName)s.',
            RoleName=details.role)['AttachedPolicies']
    except TypeError:
        LOG.info('Role %s not found.', details.role)

    for policy in attached_role_policies:
        resource_action(
            client,
            action='detach_role_policy',
            log_format='Detached Policy from Role: '
            '%(PolicyArn)s ~> %(RoleName)s',
            RoleName=details.role,
            PolicyArn=policy['PolicyArn'])

    resource_action(client, action='delete_role', log_format='Destroyed Role: %(RoleName)s', RoleName=details.role)



"""Test S3 Deploy"""

import pytest
from unittest import mock

from foremast import s3

@pytest.fixture
@mock.patch('foremast.s3.s3deploy.get_properties')
@mock.patch('foremast.s3.s3deploy.get_details')
def s3deployment(mock_get_details, mock_get_props):
    """Creates S3Deployment Fixture"""
    mock_get_props.return_value = {"deploy_strategy": "highlander",
                                   "s3": {"path": "/"}}
    mock_get_details.return_value.s3_app_bucket.return_value = "testapp"
    deployobj = s3.S3Deployment(app="testapp",
                                env="dev",
                                region="us-east-1",
                                prop_path="/",
                                artifact_path="/artifact",
                                artifact_branch="master",
                                artifact_version="1")
    return deployobj

def test_get_cmd(s3deployment):
    """Tests s3.S3Deployment._get_upload_cmd returns correct cmd"""
    expected_nomirror_cmd = "aws s3 sync /artifact s3://testapp/1 --delete --exact-timestamps --profile dev"
    expected_mirror_cmd = "aws s3 sync /artifact s3://testapp/ --delete --exact-timestamps --profile dev"
    actual_nomirror_cmd = s3deployment._get_upload_cmd(deploy_strategy="highlander")
    actual_mirror_cmd = s3deployment._get_upload_cmd(deploy_strategy="mirror")
    assert actual_nomirror_cmd == expected_nomirror_cmd
    assert actual_mirror_cmd == expected_mirror_cmd

def test_path_formatter(s3deployment):
    """Tests s3.S3Deployment._path_formatter returns correct path"""
    expected_latest_path = "s3://testapp/LATEST"
    expected_mirror_path = "s3://testapp/"
    actual_latest_path = s3deployment._path_formatter("LATEST")
    actual_mirror_path = s3deployment._path_formatter("MIRROR")
    assert actual_latest_path == expected_latest_path
    assert actual_mirror_path == expected_mirror_path



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""CLI entrypoint to Application Configuration preparer.

Help: ``python -m src.foremast.configs -h``
"""
import argparse
import logging

import gogoutils

from ..args import add_debug
from ..consts import APP_FORMATS, LOGGING_FORMAT
from .outputs import write_variables
from .prepare_configs import process_git_configs, process_runway_configs

LOG = logging.getLogger(__name__)


def main():
    """Append Application Configurations to a given file in multiple formats."""
    logging.basicConfig(format=LOGGING_FORMAT)

    parser = argparse.ArgumentParser(description=main.__doc__)
    add_debug(parser)
    parser.add_argument('-o', '--output', required=True, help='Name of environment file to append to')
    parser.add_argument(
        '-g', '--git-short', metavar='GROUP/PROJECT', required=True, help='Short name for Git, e.g. forrest/core')
    parser.add_argument('-r', '--runway-dir', help='Runway directory with app.json files, requires --git-short')
    args = parser.parse_args()

    LOG.setLevel(args.debug)
    logging.getLogger(__package__.split('.')[0]).setLevel(args.debug)

    generated = gogoutils.Generator(*gogoutils.Parser(args.git_short).parse_url(), formats=APP_FORMATS)
    git_short = generated.gitlab()['main']

    if args.runway_dir:
        configs = process_runway_configs(runway_dir=args.runway_dir)
    else:
        configs = process_git_configs(git_short=git_short)

    write_variables(app_configs=configs, out_file=args.output, git_short=git_short)


if __name__ == '__main__':
    main()



"""Verify AWS Data Pipeline Creation."""
import copy
from unittest import mock

import pytest

from foremast.datapipeline.datapipeline import AWSDataPipeline
from foremast.exceptions import DataPipelineDefinitionError

GOOD_DEF = {"objects": [], "parameters": [], "values": {}}

BAD_DEF = {"no_object": [], "parameters": [], "values": {}}

TEST_PROPERTIES = {
    'datapipeline': {
        'name': 'Test Pipeline',
        'description': "this is a test",
        'json_definition': GOOD_DEF
    }
}


@mock.patch('foremast.datapipeline.datapipeline.boto3.Session.client')
@mock.patch('foremast.datapipeline.datapipeline.get_details')
@mock.patch('foremast.datapipeline.datapipeline.get_properties')
def test_create_datapipeline(mock_get_properties, mock_get_details, mock_boto3):
    """Check data pipeline creation"""
    generated = {"project": "test"}
    properties = copy.deepcopy(TEST_PROPERTIES)
    mock_get_details.return_value.data = generated
    mock_get_properties.return_value = properties
    mock_boto3.return_value.create_pipeline.return_value = {'pipelineId': '1234'}

    dp = AWSDataPipeline(app='test_app', env='test_env', region='us-east-1', prop_path='other')
    dp.create_datapipeline()
    assert dp.pipeline_id == '1234'


@mock.patch('foremast.datapipeline.datapipeline.boto3.Session.client')
@mock.patch('foremast.datapipeline.datapipeline.get_details')
@mock.patch('foremast.datapipeline.datapipeline.get_properties')
def test_good_set_pipeline_definition(mock_get_properties, mock_get_details, mock_boto3):
    """Tests that good pipeline definition is set correctly"""
    generated = {"project": "test"}
    properties = copy.deepcopy(TEST_PROPERTIES)
    mock_get_details.return_value.data = generated
    mock_get_properties.return_value = properties

    good_dp = AWSDataPipeline(app='test_app', env='test_env', region='us-east-1', prop_path='other')
    good_dp.pipeline_id = '1'
    assert good_dp.set_pipeline_definition()


@mock.patch('foremast.datapipeline.datapipeline.boto3.Session.client')
@mock.patch('foremast.datapipeline.datapipeline.get_details')
@mock.patch('foremast.datapipeline.datapipeline.get_properties')
def test_bad_set_pipeline_definition(mock_get_properties, mock_get_details, mock_boto3):
    """Tests that bad pipeline definition is caught"""
    generated = {"project": "test"}
    properties = copy.deepcopy(TEST_PROPERTIES)
    properties['datapipeline']['json_definition'] = BAD_DEF
    mock_get_details.return_value.data = generated
    mock_get_properties.return_value = properties

    bad_dp = AWSDataPipeline(app='test_app', env='test_env', region='us-east-1', prop_path='other')
    bad_dp.pipeline_id = '1'
    with pytest.raises(DataPipelineDefinitionError):
        bad_dp.set_pipeline_definition()


@mock.patch('foremast.datapipeline.datapipeline.boto3.Session.client')
@mock.patch('foremast.datapipeline.datapipeline.get_details')
@mock.patch('foremast.datapipeline.datapipeline.get_properties')
def test_get_pipeline_id(mock_get_properties, mock_get_details, mock_boto3):
    """Tests getting the pipeline ID from boto3"""
    test_pipelines = [{
        'pipelineIdList': [{
            "name": "Test Pipeline",
            "id": "1234"
        }, {
            "name": "Other",
            "id": "5678"
        }],
        "hasMoreResults": False
    }]
    generated = {"project": "test"}
    properties = copy.deepcopy(TEST_PROPERTIES)
    mock_get_details.return_value.data = generated
    mock_get_properties.return_value = properties
    mock_boto3.return_value.get_paginator.return_value.paginate.return_value = test_pipelines

    dp = AWSDataPipeline(app='test_app', env='test_env', region='us-east-1', prop_path='other')
    dp.get_pipeline_id()
    assert dp.pipeline_id == '1234'



"""Package version functions."""
import pkg_resources


def get_version():
    """Retrieve package version."""
    version = 'Not installed.'

    try:
        version = pkg_resources.get_distribution(__package__).version
    except pkg_resources.DistributionNotFound:
        pass

    return version


def print_version():
    """Show package version."""
    print(get_version())



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create cloudwatch events"""

import json
import logging

import boto3

from ...exceptions import InvalidEventConfiguration
from ...utils import add_lambda_permissions, get_env_credential, get_lambda_arn

LOG = logging.getLogger(__name__)


def create_cloudwatch_event(app_name, env, region, rules):
    """Create cloudwatch event for lambda from rules.

    Args:
        app_name (str): name of the lambda function
        env (str): Environment/Account for lambda function
        region (str): AWS region of the lambda function
        rules (dict): Trigger rules from the settings
    """
    session = boto3.Session(profile_name=env, region_name=region)
    cloudwatch_client = session.client('events')

    rule_name = rules.get('rule_name')
    rule_type = rules.get('rule_type', 'schedule')
    schedule = rules.get('schedule')
    event_pattern = rules.get('event_pattern')
    rule_description = rules.get('rule_description')
    json_input = rules.get('json_input', {})

    if rule_type == 'schedule' and schedule is None:
        LOG.critical('A CloudWatch Schedule is required and no schedule pattern is defined!')
        raise InvalidEventConfiguration('A CloudWatch Schedule is required and no schedule is defined!')

    if rule_type == 'event_pattern' and event_pattern is None:
        LOG.critical('A CloudWatch Event Pattern is required and no event pattern is defined!')
        raise InvalidEventConfiguration('A CloudWatch Event Pattern is required and no event pattern is defined!')

    if rule_name is None:
        LOG.critical('Rule name is required and no rule_name is defined!')
        raise InvalidEventConfiguration('Rule name is required and no rule_name is defined!')
    else:
        LOG.info('%s and %s', app_name, rule_name)
        rule_name = "{}_{}".format(app_name, rule_name.replace(' ', '_'))

    if rule_description is None:
        rule_description = "{} - {}".format(app_name, rule_name)

    lambda_arn = get_lambda_arn(app=app_name, account=env, region=region)

    # Add lambda permissions
    account_id = get_env_credential(env=env)['accountId']
    principal = "events.amazonaws.com"
    statement_id = 'cloudwatch_{}'.format(rule_name)
    source_arn = 'arn:aws:events:{}:{}:rule/{}'.format(region, account_id, rule_name)
    add_lambda_permissions(
        function=lambda_arn,
        statement_id=statement_id,
        action='lambda:InvokeFunction',
        principal=principal,
        source_arn=source_arn,
        env=env,
        region=region)

    # Create CloudWatch rule
    if rule_type == 'schedule':
        cloudwatch_client.put_rule(
            Name=rule_name,
            ScheduleExpression=schedule,
            State='ENABLED',
            Description=rule_description)
        LOG.info('Created CloudWatch Rule "%s" with %s: %s', rule_name, rule_type, schedule)
    elif rule_type == 'event_pattern':
        cloudwatch_client.put_rule(
            Name=rule_name,
            EventPattern=json.dumps(event_pattern),
            State='ENABLED',
            Description=rule_description)
        LOG.info('Created CloudWatch Rule "%s" with %s: %s', rule_name, rule_type, event_pattern)

    targets = [{
        "Id": app_name,
        "Arn": lambda_arn,
    }]

    if json_input:
        json_payload = '{}'.format(json.dumps(json_input))
        for each_target in targets:
            each_target['Input'] = json_payload

    put_targets_response = cloudwatch_client.put_targets(Rule=rule_name, Targets=targets)
    LOG.debug('CloudWatch PutTargets Response: %s', put_targets_response)
    LOG.info('Configured CloudWatch Rule Target: %s', lambda_arn)



"""Test Stage renumerate logic."""
from foremast.pipeline.renumerate_stages import renumerate_stages


def test_basic():
    """Should be sequential Stage IDs.
    Examples:
        Graphically:

        .. code-block:: text

            [ 1 ]-[ 2 ]-[ 3 ]
    """
    pipeline = {
        'stages': [
            {
                'refId': 'master',
            },
            {
                'refId': 'master',
            },
            {
                'refId': 'master',
            },
        ],
    }

    answer = {
        'stages': [
            {
                'refId': '1',
                'requisiteStageRefIds': [],
            },
            {
                'refId': '2',
                'requisiteStageRefIds': ['1'],
            },
            {
                'refId': '3',
                'requisiteStageRefIds': ['2'],
            },
        ],
    }

    assert renumerate_stages(pipeline) == answer


def test_branches():
    """Test 'branch' Stages refer to correct 'master' level.

    Examples:
        Graphically:

        .. code-block:: text

            [ 1 ]-[ 2 ]-[ 3 ]-[ 4 ]
                    L---[200]
                    L---[201]
                    L---[202]
    """
    pipeline = {
        'stages': [
            {
                'refId': 'master',
            },
            {
                'refId': 'master',
            },
            {
                'refId': 'branch',
            },
            {
                'refId': 'branch',
            },
            {
                'refId': 'branch',
            },
            {
                'refId': 'master',
            },
            {
                'refId': 'master',
            },
        ],
    }

    answer = {
        'stages': [
            {
                'refId': '1',
                'requisiteStageRefIds': [],
            },
            {
                'refId': '2',
                'requisiteStageRefIds': ['1'],
            },
            {
                'refId': '200',
                'requisiteStageRefIds': ['2'],
            },
            {
                'refId': '201',
                'requisiteStageRefIds': ['2'],
            },
            {
                'refId': '202',
                'requisiteStageRefIds': ['2'],
            },
            {
                'refId': '3',
                'requisiteStageRefIds': ['2'],
            },
            {
                'refId': '4',
                'requisiteStageRefIds': ['3'],
            },
        ],
    }

    assert renumerate_stages(pipeline) == answer



"""SNS Topic functions."""
import logging

import boto3

from ..exceptions import SNSTopicNotFound

LOG = logging.getLogger(__name__)


def get_sns_topic_arn(topic_name, account, region):
    """Get SNS topic ARN.

    Args:
        topic_name (str): Name of the topic to lookup.
        account (str): Environment, e.g. dev
        region (str): Region name, e.g. us-east-1

    Returns:
        str: ARN for requested topic name

    """
    if topic_name.count(':') == 5 and topic_name.startswith('arn:aws:sns:'):
        return topic_name
    session = boto3.Session(profile_name=account, region_name=region)
    sns_client = session.client('sns')

    topics = sns_client.list_topics()['Topics']

    matched_topic = None
    for topic in topics:
        topic_arn = topic['TopicArn']
        if topic_name == topic_arn.split(':')[-1]:
            matched_topic = topic_arn
            break
    else:
        LOG.critical("No topic with name %s found.", topic_name)
        raise SNSTopicNotFound('No topic with name {0} found'.format(topic_name))
    return matched_topic



"""Test AWS Lambda Utilities."""
from unittest import mock

import boto3
import pytest

from foremast.exceptions import LambdaAliasDoesNotExist
from foremast.utils.awslambda import add_lambda_permissions, get_lambda_alias_arn

ERROR_RESPONSE = {'Error': {}}


def lambda_alias_list_mock():
    return_dict = {
        'Aliases': [{
            'Description': '',
            'AliasArn': 'arn:aws:lambda:us-east-1:222572804561:function:lambdatest2:stage',
            'FunctionVersion': '$LATEST',
            'Name': 'stage'
        }, {
            'Description': '',
            'AliasArn': 'arn:aws:lambda:us-east-1:222572804561:function:lambdatest2:dev',
            'FunctionVersion': '$LATEST',
            'Name': 'dev'
        }]
    }
    return return_dict


def lambda_no_alias_list_mock():
    return_dict = {'Aliases': []}
    return return_dict


@mock.patch('foremast.utils.awslambda.boto3')
@mock.patch('foremast.utils.awslambda.LOG')
def test_add_lambda_permission_success(LOG, mock_boto3):
    """Check lambda permssion add successful."""
    add_lambda_permissions()

    args, _ = LOG.info.call_args
    assert args[0].startswith('Add permission')


@mock.patch('foremast.utils.awslambda.boto3.Session')
@mock.patch('foremast.utils.awslambda.LOG')
def test_add_lambda_permission_failure(LOG, session):
    """Check lambda permssion add failure."""
    client = session.return_value.client.return_value
    client.add_permission.side_effect = boto3.exceptions.botocore.exceptions.ClientError(
        ERROR_RESPONSE, 'operation_name')

    add_lambda_permissions()

    args, _ = LOG.info.call_args
    assert args[0].startswith('Did not add')


@mock.patch('foremast.utils.awslambda.boto3.Session')
@mock.patch('foremast.utils.awslambda.LOG')
def test_get_lambda_alias_arn_success(LOG, mock_boto3):
    """Check get lambda alias arn."""
    client = mock_boto3.return_value.client.return_value
    client.list_aliases.return_value = lambda_alias_list_mock()

    rv = get_lambda_alias_arn('lambdatest', 'dev', 'us-east-1')

    args, _ = LOG.info.call_args

    assert rv == 'arn:aws:lambda:us-east-1:222572804561:function:lambdatest2:dev'
    assert args[0].startswith('Found ARN for alias')


@mock.patch('foremast.utils.awslambda.boto3.Session')
def test_get_lambda_alias_arn_failure(mock_boto3):
    """Check get lambda alias arn failure."""
    client = mock_boto3.return_value.client.return_value
    client.list_aliases.return_value = lambda_no_alias_list_mock()

    with pytest.raises(LambdaAliasDoesNotExist):
        get_lambda_alias_arn('lambdatest', 'dev', 'us-east-1')



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create cloudwatch log event"""

import logging

import boto3

from ...exceptions import InvalidEventConfiguration
from ...utils import add_lambda_permissions, get_env_credential, get_lambda_alias_arn

LOG = logging.getLogger(__name__)


def create_cloudwatch_log_event(app_name, env, region, rules):
    """Create cloudwatch log event for lambda from rules.

    Args:
        app_name (str): name of the lambda function
        env (str): Environment/Account for lambda function
        region (str): AWS region of the lambda function
        rules (str): Trigger rules from the settings
    """

    session = boto3.Session(profile_name=env, region_name=region)
    cloudwatch_client = session.client('logs')

    log_group = rules.get('log_group')
    filter_name = rules.get('filter_name')
    filter_pattern = rules.get('filter_pattern')

    if not log_group:
        LOG.critical('Log group is required and no "log_group" is defined!')
        raise InvalidEventConfiguration('Log group is required and no "log_group" is defined!')

    if not filter_name:
        LOG.critical('Filter name is required and no filter_name is defined!')
        raise InvalidEventConfiguration('Filter name is required and no filter_name is defined!')

    if filter_pattern is None:
        LOG.critical('Filter pattern is required and no filter_pattern is defined!')
        raise InvalidEventConfiguration('Filter pattern is required and no filter_pattern is defined!')

    lambda_alias_arn = get_lambda_alias_arn(app=app_name, account=env, region=region)

    statement_id = '{}_cloudwatchlog_{}'.format(app_name, filter_name.replace(" ", "_"))
    principal = 'logs.{}.amazonaws.com'.format(region)
    account_id = get_env_credential(env=env)['accountId']
    source_arn = "arn:aws:logs:{0}:{1}:log-group:{2}:*".format(region, account_id, log_group)
    add_lambda_permissions(
        function=lambda_alias_arn,
        statement_id=statement_id,
        action='lambda:InvokeFunction',
        principal=principal,
        source_arn=source_arn,
        env=env,
        region=region)

    cloudwatch_client.put_subscription_filter(
        logGroupName=log_group, filterName=filter_name, filterPattern=filter_pattern, destinationArn=lambda_alias_arn)

    LOG.info("Created Cloudwatch log event with filter: %s", filter_pattern)



#   Foremast - Pipeline Tooling
#
#   Copyright 2020 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Retrieve GCP Environments and their configuration"""

from ..consts import GCP_ENVS
from ..exceptions import GoogleInfrastructureError
from google.oauth2 import service_account
from googleapiclient import discovery
import logging

LOG = logging.getLogger(__name__)


class GcpEnvironment:

    def __init__(self, name, **entries):
        """Creates a new GcpEnvironment object
        Args:
            name: str, the name of the environment
            entries: Dict of properties to be serialized into the class
        """
        self.name = name
        self.organization = None
        self._all_projects_cache = []
        self._single_project_cache = dict()
        self.service_account_project = None
        self.service_account_path = None
        self.__dict__.update(entries)

    def get_credentials(self):
        """Gets a GCP service account credentials"""
        credentials = service_account.Credentials.from_service_account_file(
            filename=self.service_account_path,
            scopes=['https://www.googleapis.com/auth/cloud-platform'])
        return credentials

    def get_all_projects(self):
        """Returns all projects in this environment.  If this method is called more
        than once a cached response will be returned to avoid duplicate calls to GCP APIs."""

        # Check cache for all projects in this env to avoid duplicate API calls to GCP
        if self._all_projects_cache:
            LOG.debug("Reusing GCP projects cache for environment {}".format(self.name))
            return self._all_projects_cache

        # No cached response found, check GCP APIs
        service = discovery.build('cloudresourcemanager', 'v1', credentials=self.get_credentials(),
                                  cache_discovery=False)
        project_filter = self._get_project_api_filter()
        request = service.projects().list(filter=project_filter)
        response = request.execute()
        projects = response.get('projects', [])

        if not projects:
            raise GoogleInfrastructureError("No projects returned for filter {}. ".format(project_filter)
                                            + "A Foremast GCP Environment needs at least one project.")

        self._all_projects_cache = projects

        return projects

    def get_project(self, project_name):
        """Gets the project for the given project prefix in this environment.
        If duplicate calls to this method with the same arguments are made
        a cached response will be used to avoid duplicate calls to GCP APIs"""

        # Check cache for this project in this env to avoid duplicate API calls to GCP
        if project_name in self._single_project_cache:
            LOG.debug("Reusing GCP projects cache for environment {} and project {}".format(self.name, project_name))
            return self._single_project_cache[project_name]

        # No cached response found, check GCP APIs
        service = discovery.build('cloudresourcemanager', 'v1', credentials=self.get_credentials())
        project_filter = self._get_project_api_filter(name=project_name)
        request = service.projects().list(filter=project_filter)
        response = request.execute()
        projects = response.get('projects', [])

        # No projects found
        if not projects:
            error_message = "No projects returned for filter {}".format(project_filter)
            raise GoogleInfrastructureError(error_message)

        # If more than one project is found, Foremast cannot determine which project should be used
        # raise and error with the projects so the user can adjust their labels or project_name
        if len(projects) > 1:
            error_message = "More than one project returned for filter {}. Projects returned:".format(project_filter)
            for project in projects:
                error_message += " " + project['name']
            raise GoogleInfrastructureError(error_message)

        project = projects[0]
        self._single_project_cache[project_name] = project
        return project

    def _get_project_api_filter(self, name=None):
        """Gets the project filter based on the name and env for use in Google IAM APIs"""
        base_filter = "labels.cloud_env:{} labels.foremast_enabled=true".format(self.name)

        if name is not None:
            base_filter += " name:{}".format(name)

        return base_filter

    @staticmethod
    def get_environments_from_config():
        gcp_envs = dict()
        for env_name in GCP_ENVS:
            env_config = GCP_ENVS[env_name]
            gcp_envs[env_name] = GcpEnvironment(name=env_name, **env_config)
        return gcp_envs



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Search for ELB DNS name."""
import logging

import boto3
from tryagain import retries

from ..exceptions import SpinnakerElbNotFound
from ..utils.gate import gate_request

LOG = logging.getLogger(__name__)


@retries(max_attempts=5, wait=2, exceptions=(AssertionError, SpinnakerElbNotFound))
def find_elb(name='', env='', region=''):
    """Get an application's AWS elb dns name.

    Args:
        name (str): ELB name
        env (str): Environment/account of ELB
        region (str): AWS Region

    Returns:
        str: elb DNS record

    """
    LOG.info('Find %s ELB in %s [%s].', name, env, region)

    uri = '/applications/{0}/loadBalancers'.format(name)
    response = gate_request(uri=uri)
    assert response.ok

    elb_dns = None
    accounts = response.json()
    for account in accounts:
        if account['account'] == env and account['region'] == region:
            elb_dns = account['dnsname']
            break
    else:
        raise SpinnakerElbNotFound('Elb for "{0}" in region {1} not found'.format(name, region))

    LOG.info('Found: %s', elb_dns)
    return elb_dns


def find_elb_dns_zone_id(name='', env='dev', region='us-east-1'):
    """Get an application's AWS elb dns zone id.

    Args:
        name (str): ELB name
        env (str): Environment/account of ELB
        region (str): AWS Region

    Returns:
        str: elb DNS zone ID

    """
    LOG.info('Find %s ELB DNS Zone ID in %s [%s].', name, env, region)
    client = boto3.Session(profile_name=env).client('elb', region_name=region)
    elbs = client.describe_load_balancers(LoadBalancerNames=[name])
    return elbs['LoadBalancerDescriptions'][0]['CanonicalHostedZoneNameID']



#   Foremast - Pipeline Tooling
#
#   Copyright 2021 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create AWS Step Functions"""

import logging

import json
import boto3

from ..utils import get_details, get_properties, get_role_arn

LOG = logging.getLogger(__name__)


class AWSStepFunction:
    """Manipulate Step Function."""

    def __init__(self, app=None, env=None, region='us-east-1', prop_path=None):
        """AWS Step Function object.

        Args:
            app (str): Application name
            env (str): Environment/Account
            region (str): AWS Region
            prop_path (str): Path of environment property file
        """
        self.app_name = app
        self.env = env
        self.region = region
        self.properties = get_properties(prop_path, env=self.env, region=self.region)
        self.stepfunction_data = self.properties['stepfunction']
        self.custom_tags = self.properties['app']['custom_tags']
        generated = get_details(app=self.app_name)
        self.group = generated.data['project']

        session = boto3.Session(profile_name=self.env, region_name=self.region)
        self.client = session.client('stepfunctions')

        self.state_machine_arn = None
        self.role = generated.iam()['role']
        self.role_arn = get_role_arn(self.role, self.env, self.region)

    def create_stepfunction(self):
        """Creates the Step Function if it does not already exist

        Returns:
                dict: the response of the Boto3 command
        """

        stepfunction_tags = [
            {'key': 'app_group', 'value': self.group},
            {'key': 'app_name', 'value': self.app_name}
        ]

        for each_tag in self.custom_tags:
            stepfunction_tags.append({'key': each_tag, 'value': self.custom_tags[each_tag]})

        self.role_arn = get_role_arn(self.role, self.env, self.region)

        if not self.find_stepfunction_arn():
            response = self.client.create_state_machine(
                name=self.stepfunction_data.get('name', self.app_name),
                definition=json.dumps(self.stepfunction_data['json_definition']),
                roleArn=self.role_arn,
                type=self.stepfunction_data['statemachine_type'],
                loggingConfiguration=self.stepfunction_data['logging_configuration'],
                tags=stepfunction_tags,
                tracingConfiguration=self.stepfunction_data['tracing'])
            LOG.debug(response)

            self.state_machine_arn = response.get('stateMachineArn')

            LOG.info("Successfully configured Step Function - %s", self.app_name)
        else:
            response = self.update_stepfunction_definition()

        return response

    def update_stepfunction_definition(self):
        """Translates the json definition and puts it on created Step Function

        Returns:
                dict: the response of the Boto3 command
        """

        stepfunction_tags = [
            {'key': 'app_group', 'value': self.group},
            {'key': 'app_name', 'value': self.app_name}
        ]

        for each_tag in self.custom_tags:
            stepfunction_tags.append({'key': each_tag, 'value': self.custom_tags[each_tag]})

        if not self.state_machine_arn:
            self.find_stepfunction_arn()

        response = self.client.update_state_machine(
            stateMachineArn=self.state_machine_arn,
            definition=json.dumps(self.stepfunction_data['json_definition']),
            roleArn=self.role_arn,
            loggingConfiguration=self.stepfunction_data['logging_configuration'],
            tracingConfiguration=self.stepfunction_data['tracing'])
        LOG.debug(response)
        LOG.info("Successfully updated Step Function State Machine definition")

        LOG.info('Updating Step Function tags')

        self.client.tag_resource(resourceArn=self.state_machine_arn, tags=stepfunction_tags)

        return response

    def find_stepfunction_arn(self):
        """Finds the Step Function for configured pipeline"""

        all_statemachines = []
        paginiator = self.client.get_paginator('list_state_machines')
        for page in paginiator.paginate():
            all_statemachines.extend(page['stateMachines'])

        for statemachine in all_statemachines:
            if statemachine['name'] == self.stepfunction_data.get('name', self.app_name):
                self.state_machine_arn = statemachine['stateMachineArn']
                LOG.info("Step Function State Machine Found")
                return True
        LOG.info("Step Function State Machine Not Found for %s", self.app_name)
        return False



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Destroy S3 events."""
import logging

import boto3

from ....utils import get_details

LOG = logging.getLogger(__name__)


def destroy_s3_event(app, env, region):
    """Destroy S3 event.

    Args:
        app (str): Spinnaker Application name.
        env (str): Deployment environment.
        region (str): AWS region.
    Returns:
        bool: True upon successful completion.
    """

    # TODO: how do we know which bucket to process if triggers dict is empty?
    # Maybe list buckets and see which has notification to that lambda defined?
    # TODO: buckets should be named the same as apps, what if one app has multiple buckets?
    # bucket = rules.get('bucket')
    generated = get_details(app=app, env=env)

    bucket = generated.s3_app_bucket()

    session = boto3.Session(profile_name=env, region_name=region)
    s3_client = session.client('s3')

    config = {}

    s3_client.put_bucket_notification_configuration(Bucket=bucket, NotificationConfiguration=config)
    LOG.debug("Deleted Lambda S3 notification")

    return True



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create SNS event for lambda"""

import logging

import boto3

from ...utils import add_lambda_permissions, get_lambda_alias_arn, get_sns_topic_arn

LOG = logging.getLogger(__name__)


def create_sns_event(app_name, env, region, rules):
    """Create SNS lambda event from rules.

    Args:
        app_name (str): name of the lambda function
        env (str): Environment/Account for lambda function
        region (str): AWS region of the lambda function
        rules (str): Trigger rules from the settings
    """
    session = boto3.Session(profile_name=env, region_name=region)
    sns_client = session.client('sns')

    topic_name = rules.get('topic')
    lambda_alias_arn = get_lambda_alias_arn(app=app_name, account=env, region=region)
    topic_arn = get_sns_topic_arn(topic_name=topic_name, account=env, region=region)
    protocol = 'lambda'

    statement_id = '{}_sns_{}'.format(app_name, topic_name)
    principal = 'sns.amazonaws.com'
    add_lambda_permissions(
        function=lambda_alias_arn,
        statement_id=statement_id,
        action='lambda:InvokeFunction',
        principal=principal,
        source_arn=topic_arn,
        env=env,
        region=region)

    sns_client.subscribe(TopicArn=topic_arn, Protocol=protocol, Endpoint=lambda_alias_arn)
    LOG.debug("SNS Lambda event created")

    LOG.info("Created SNS event subscription on topic %s", topic_name)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Construct a block section of Stages in a Spinnaker Pipeline."""
import collections
import copy
import json
import logging
from pprint import pformat

from ..consts import ASG_WHITELIST, DEFAULT_EC2_SECURITYGROUPS, EC2_PIPELINE_TYPES, ENV_CONFIGS
from ..utils import generate_encoded_user_data, get_template, remove_duplicate_sg, verify_approval_skip

LOG = logging.getLogger(__name__)


def check_provider_healthcheck(settings, default_provider='Discovery'):
    """Set Provider Health Check when specified.

    Returns:
        collections.namedtuple: **ProviderHealthCheck** with attributes:

            * providers (list): Providers set to use native Health Check.
            * has_healthcheck (bool): If any native Health Checks requested.
    """
    ProviderHealthCheck = collections.namedtuple('ProviderHealthCheck', ['providers', 'has_healthcheck'])

    eureka_enabled = settings['app']['eureka_enabled']
    providers = settings['asg']['provider_healthcheck']

    LOG.debug('Template defined Health Check Providers: %s', providers)

    health_check_providers = []
    has_healthcheck = False

    normalized_default_provider = default_provider.capitalize()

    if eureka_enabled:
        LOG.info('Eureka enabled, enabling default Provider Health Check: %s', normalized_default_provider)

        for provider, active in providers.items():
            if provider.lower() == normalized_default_provider.lower():
                providers[provider] = True
                LOG.debug('Override defined Provider Health Check: %s -> %s', active, providers[provider])
                break
        else:
            LOG.debug('Adding default Provider Health Check: %s', normalized_default_provider)
            providers[normalized_default_provider] = True

    for provider, active in providers.items():
        if active:
            health_check_providers.append(provider.capitalize())
    LOG.info('Provider healthchecks: %s', health_check_providers)

    if health_check_providers:
        has_healthcheck = True

    return ProviderHealthCheck(providers=health_check_providers, has_healthcheck=has_healthcheck)


def get_template_name(env, pipeline_type):
    """Generates the correct template name based on pipeline type

    Args:
        env (str): environment to generate templates for
        pipeline_type (str): Type of pipeline like ec2 or lambda

    Returns:
        str: Name of template
    """
    pipeline_base = 'pipeline/pipeline'
    template_name_format = '{pipeline_base}'
    if env.startswith('prod'):
        template_name_format = template_name_format + '_{env}'
    else:
        template_name_format = template_name_format + '_stages'

    if pipeline_type != 'ec2':
        template_name_format = template_name_format + '_{pipeline_type}'

    template_name_format = template_name_format + '.json.j2'
    template_name = template_name_format.format(pipeline_base=pipeline_base, env=env, pipeline_type=pipeline_type)

    return template_name


def construct_pipeline_block(env='',
                             generated=None,
                             previous_env='',
                             region='us-east-1',
                             settings=None,
                             pipeline_data=None,
                             region_subnets=None,
                             **kwargs):
    """Create the Pipeline JSON from template.

    This handles the common repeatable patterns in a pipeline, such as
    judgement, infrastructure, tagger and qe.

    Note:
       ASG Health Check type is overridden to `EC2` when deploying to **dev** or
       using :ref:`eureka_enabled`.

    Args:
        env (str): Deploy environment name, e.g. dev, stage, prod.
        generated (gogoutils.Generator): Gogo Application name generator.
        kwargs (dict): Extra variables to pass to Pipeline Templates.
        previous_env (str): The previous deploy environment to use as
            Trigger.
        region (str): AWS Region to deploy to.
        settings (dict): Environment settings from configurations.
        pipeline_data (dict): Pipeline settings from configurations
        region_subnets (dict): Subnets for a Region, e.g.
            {'us-west-2': ['us-west-2a', 'us-west-2b', 'us-west-2c']}.

    Returns:
        dict: Pipeline JSON template rendered with configurations.

    """
    LOG.info('%s block for [%s].', env, region)
    LOG.debug('%s info:\n%s', env, pformat(settings))

    pipeline_type = pipeline_data['type']

    if pipeline_type in EC2_PIPELINE_TYPES:
        data = ec2_pipeline_setup(
            generated=generated,
            settings=settings,
            env=env,
            region=region,
            pipeline_type=pipeline_type,
            project=generated.project,
            region_subnets=region_subnets,
        )
    else:
        data = copy.deepcopy(settings)

    approval_skip = verify_approval_skip(data, env, ENV_CONFIGS)

    data['app'].update({
        'appname': generated.app_name(),
        'approval_skip': approval_skip,
        'repo_name': generated.repo,
        'group_name': generated.project,
        'environment': env,
        'region': region,
        'previous_env': previous_env,
        'promote_restrict': pipeline_data['promote_restrict'],
        'owner_email': pipeline_data['owner_email'],
        'pipeline': pipeline_data,
    })

    LOG.debug('Block data:\n%s', pformat(data))

    template_name = get_template_name(env, pipeline_type)
    pipeline_json = get_template(template_file=template_name, data=data, formats=generated, **kwargs)
    return pipeline_json


def ec2_pipeline_setup(
        generated=None,
        project='',
        settings=None,
        env='',
        pipeline_type='',
        region='',
        region_subnets=None,
):
    """Handles ec2 pipeline data setup

    Args:
        generated (gogoutils.Generator): Generated naming formats.
        project (str): Group name of application
        settings (dict): Environment settings from configurations.
        env (str): Deploy environment name, e.g. dev, stage, prod.
        pipeline_type (str): Type of Foremast Pipeline to configure.
        region (str): AWS Region to deploy to.
        region_subnets (dict): Subnets for a Region, e.g.
            {'us-west-2': ['us-west-2a', 'us-west-2b', 'us-west-2c']}.

    Returns:
        dict: Updated settings to pass to templates for EC2 info

    """
    data = copy.deepcopy(settings)
    user_data = generate_encoded_user_data(
        env=env,
        region=region,
        generated=generated,
        group_name=project,
        pipeline_type=pipeline_type,
    )

    # Use different variable to keep template simple
    instance_security_groups = sorted(DEFAULT_EC2_SECURITYGROUPS[env])
    instance_security_groups.append(generated.security_group_app)
    instance_security_groups.extend(settings['security_group']['instance_extras'])
    instance_security_groups = remove_duplicate_sg(instance_security_groups)

    LOG.info('Instance security groups to attach: %s', instance_security_groups)

    # check if scaling policy exists
    if settings['asg']['scaling_policy']:
        scalingpolicy = True
        LOG.info('Found scaling policy')
    else:
        scalingpolicy = False
        LOG.info('No scaling policy found')

    if settings['app']['eureka_enabled']:
        elb = []
    else:
        elb = [generated.elb_app]
    LOG.info('Attaching the following ELB: %s', elb)

    health_checks = check_provider_healthcheck(settings)

    # Use EC2 Health Check for DEV or Eureka enabled
    if env == 'dev' or settings['app']['eureka_enabled']:
        data['asg'].update({'hc_type': 'EC2'})
        LOG.info('Switching health check type to: EC2')

    # Aggregate the default grace period, plus the exposed app_grace_period
    # to allow per repo extension of asg healthcheck grace period
    hc_grace_period = data['asg'].get('hc_grace_period')
    app_grace_period = data['asg'].get('app_grace_period')
    grace_period = hc_grace_period + app_grace_period

    # TODO: Migrate the naming logic to an external library to make it easier
    #       to update in the future. Gogo-Utils looks like a good candidate
    ssh_keypair = data['asg'].get('ssh_keypair', None)
    if not ssh_keypair:
        ssh_keypair = '{0}_{1}_default'.format(env, region)
    LOG.info('SSH keypair (%s) used', ssh_keypair)

    if settings['app']['canary']:
        canary_user_data = generate_encoded_user_data(
            env=env,
            region=region,
            generated=generated,
            group_name=project,
            canary=True,
        )
        data['app'].update({
            'canary_encoded_user_data': canary_user_data,
        })

    data['asg'].update({
        'hc_type': data['asg'].get('hc_type').upper(),
        'hc_grace_period': grace_period,
        'ssh_keypair': ssh_keypair,
        'provider_healthcheck': json.dumps(health_checks.providers),
        'enable_public_ips': json.dumps(settings['asg']['enable_public_ips']),
        'has_provider_healthcheck': health_checks.has_healthcheck,
        'asg_whitelist': ASG_WHITELIST,
    })

    data['app'].update({
        'az_dict': json.dumps(region_subnets),
        'encoded_user_data': user_data,
        'instance_security_groups': json.dumps(instance_security_groups),
        'elb': json.dumps(elb),
        'scalingpolicy': scalingpolicy,
    })

    return data



"""Spinnaker validate functions."""
import logging

from .consts import API_URL
from .utils.credentials import get_env_credential

LOG = logging.getLogger(__name__)


def validate_gate():
    """Check Gate connection."""
    try:
        credentials = get_env_credential()
        LOG.debug('Found credentials: %s', credentials)
        LOG.info('Gate working.')
    except TypeError:
        LOG.fatal('Gate connection not valid: API_URL = %s', API_URL)


def validate_all(args):
    """Run all validate steps."""
    LOG.debug('Args: %s', args)

    LOG.info('Running all validate steps.')
    validate_gate()



#   Foremast - Pipeline Tooling
#
#   Copyright 2020 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Test generate_predictable_pipeline_id functionality"""
from foremast.utils.pipelines import generate_predictable_pipeline_id

def test_generate_predictable_pipeline_id():
    """Tests that generate_predictable_pipeline_id creates a matching UUID with identical seed inputs"""
    app_name = "testspinnakerapplication"
    pipeline_name = "testspinnakerpipeline"
    uuid_1 = generate_predictable_pipeline_id(app_name, pipeline_name)
    uuid_2 = generate_predictable_pipeline_id(app_name, pipeline_name)
    assert uuid_1 == uuid_2



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC #
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""POST a new task or check status of running task."""
import copy
import json
import logging
from functools import partial

from tryagain import call as retry_call

from ..consts import DEFAULT_TASK_TIMEOUT, HEADERS, TASK_TIMEOUTS
from ..exceptions import SpinnakerTaskError, SpinnakerTaskInconclusiveError
from ..utils import gate_request

LOG = logging.getLogger(__name__)


def post_task(task_data, task_uri='/tasks'):
    """Create Spinnaker Task.

    Args:
        task_data (str): Task JSON definition.

    Returns:
        str: Spinnaker Task ID.

    Raises:
        AssertionError: Error response from Spinnaker.

    """
    uri = '/{}'.format(task_uri.lstrip('/'))

    if isinstance(task_data, str):
        task_json = task_data
    else:
        task_json = json.dumps(task_data)

    resp = gate_request(method='POST', uri=uri, data=task_json, headers=HEADERS)
    resp_json = resp.json()

    LOG.debug(resp_json)

    assert resp.ok, 'Spinnaker communication error: {0}'.format(resp.text)

    return resp_json['ref']


def _check_task(taskid):
    """Check Spinnaker Task status.

    Args:
        taskid (str): Existing Spinnaker Task ID.

    Returns:
        str: Task status.

    """
    try:
        taskurl = taskid.get('ref', '0000')
    except AttributeError:
        taskurl = taskid

    taskid = taskurl.split('/tasks/')[-1]

    LOG.info('Checking taskid %s', taskid)

    headers = copy.copy(HEADERS)
    headers.pop('content-type')

    uri = '/tasks/{}'.format(taskid)
    task_response = gate_request(uri=uri, headers=headers)

    LOG.debug(task_response.json())

    assert task_response.ok, 'Spinnaker communication error: {0}'.format(task_response.text)

    task_state = task_response.json()
    status = task_state['status']
    LOG.info('Current task status: %s', status)

    if status == 'SUCCEEDED':  # pylint: disable=no-else-return
        return status
    elif status == 'TERMINAL':
        raise SpinnakerTaskError(task_state)
    else:
        raise ValueError


def check_task(taskid, timeout=DEFAULT_TASK_TIMEOUT, wait=2):
    """Wrap check_task.

    Args:
        taskid (str): Existing Spinnaker Task ID.
        timeout (int, optional): Consider Task failed after given seconds.
        wait (int, optional): Seconds to pause between polling attempts.

    Returns:
        str: Task status.

    Raises:
        AssertionError: API did not respond with a 200 status code.
        :obj:`foremast.exceptions.SpinnakerTaskInconclusiveError`: Task did not
            reach a terminal state before the given time out.

    """
    max_attempts = int(timeout / wait)
    try:
        return retry_call(
            partial(_check_task, taskid),
            max_attempts=max_attempts,
            wait=wait,
            exceptions=(AssertionError, ValueError), )
    except ValueError:
        raise SpinnakerTaskInconclusiveError('Task failed to complete in {0} seconds: {1}'.format(timeout, taskid))


def wait_for_task(task_data, task_uri='/tasks'):
    """Run task and check the result.

    Args:
        task_data (str): the task json to execute

    Returns:
        str: Task status.

    """
    taskid = post_task(task_data, task_uri)

    if isinstance(task_data, str):
        json_data = json.loads(task_data)
    else:
        json_data = task_data

    # inspect the task to see if a timeout is configured
    job = json_data['job'][0]
    env = job.get('credentials')
    task_type = job.get('type')

    timeout = TASK_TIMEOUTS.get(env, dict()).get(task_type, DEFAULT_TASK_TIMEOUT)

    LOG.debug("Task %s will timeout after %s", task_type, timeout)

    return check_task(taskid, timeout)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Test validity of runner."""
import os
from unittest import mock

import pytest

from foremast.runner import ForemastRunner
from foremast.pipeline import SpinnakerPipeline

CONFIGS = {
    'pipeline': {
        'type': None,
    },
}

ENV_VARIABLES = {
    'EMAIL': 'test@localhost.localdomain',
    'ENV': 'dev',
    'PROJECT': 'group1',
    'GIT_REPO': 'repo1',
    'REGION': 'us-east-1',
    'RUNWAY_DIR': '.',
}

for name, value in ENV_VARIABLES.items():
    os.environ[name] = value


def test_runner_create_pipeline_invalid_type():
    """Test invalid pipeline types."""
    runner = ForemastRunner()
    runner.configs = CONFIGS
    with pytest.raises(NotImplementedError):
        runner.configs['pipeline']['type'] = 'not_allowed_type'
        runner.create_pipeline()


@mock.patch('foremast.runner.consts.ALLOWED_TYPES', new=['manual'])
@mock.patch('foremast.runner.pipeline')
def test_runner_create_pipeline_onetime(mock_pipelines):
    """Test onetime pipeline."""
    runner = ForemastRunner()
    runner.configs = CONFIGS
    runner.configs['pipeline']['type'] = 'manual'
    runner.create_pipeline(onetime=True)
